"""
--- Day 20: Race Condition ---
https://adventofcode.com/2024/day/20
"""
from aocd import data
from aocd import extra

from aoc_wim.zgrid import ZGrid


grid = ZGrid(data)
S, E = grid.z("S"), grid.z("E")
grid[S] = "."
dzs = -1j, 1, 1j, -1
path = [z_prev] = [E]
while path[-1] != S:
    for dz in dzs:
        z = path[-1] + dz
        if grid[z] == "." and z != z_prev:
            z_prev = path[-1]
            path.append(z)
            break
path = {z: i for i, z in enumerate(path)}  # {pos: distance from goal}


def n_cheats(t_cheat, dt_min):
    result = 0
    for z0, d in path.items():
        for r in range(1, t_cheat + 1):
            for z in grid.zball(z0, r):
                if z in path:
                    result += d - r - path[z] >= dt_min
    return result


dt_min = extra.get("dt_min", 100)
print("answer_a:", n_cheats(2, dt_min))
print("answer_b:", n_cheats(20, dt_min))



"""
--- Day 17: Spinlock ---
https://adventofcode.com/2017/day/17
"""
from aocd import data

n = int(data)

L = [0]
pos = 0
for i in range(1, 2017 + 1):
    pos = 1 + (pos + n) % i
    L.insert(pos, i)
print("answer_a:", L[pos + 1])

pos = 0
i = b = 1
while i <= 50000000:
    pos = 1 + (pos + n) % i
    if pos == 1:
        b = i
    # when pos is much smaller than i, we can skip
    # ahead iterations. as i gets bigger, this
    # becomes an increasingly effective speedup
    inci = (i - pos) // (n + 1) or 1
    pos += (n + 1) * (inci - 1)
    i += inci
print("answer_b:", b)



"""
--- Day 4: Ceres Search ---
https://adventofcode.com/2024/day/4
"""
from aocd import data
from aoc_wim.zgrid import ZGrid; grid = ZGrid(data)


linesT = ["".join(x) for x in zip(*data.splitlines())]

a = 0
for line in data.splitlines():
    a += line.count("XMAS") + line.count("SAMX")
for line in linesT:
    a += line.count("XMAS") + line.count("SAMX")

linesD1 = linesT.copy()
linesD2 = linesT.copy()
for i, line in enumerate(linesT):
    head = " " * i
    tail = " " * (len(line) - 1 - i)
    linesD1[i] = head + line + tail
    linesD2[i] = tail + line + head
linesD1 = ["".join(line) for line in zip(*linesD1)]
linesD2 = ["".join(line) for line in zip(*linesD2)]
for line in linesD1:
    a += line.count("XMAS") + line.count("SAMX")
for line in linesD2:
    a += line.count("XMAS") + line.count("SAMX")

b = 0
for z0, g in grid.items():
    if g == "A":
        vals = "".join([grid.get(z0 + z, ".") for z in [-1j-1, -1j+1, 1j+1, 1j-1]])
        b += vals in ["MSSM", "SSMM", "MMSS", "SMMS"]



print("answer_a:", a)
print("answer_b:", b)



"""
--- Day 8: Memory Maneuver ---
https://adventofcode.com/2018/day/8
"""
from anytree import NodeMixin
from aocd import data


class MyNode(NodeMixin):
    def __init__(self, parent=None):
        self.parent = parent

    @property
    def val(self):
        children = self.children
        if len(children) == 0:
            return sum(self.metas)
        else:
            total = 0
            for m in self.metas:
                try:
                    total += children[m - 1].val
                except IndexError:
                    pass
            return total


ns = [int(x) for x in data.split()]
sum_all_metas = 0
node = root = MyNode()
try:
    i = 0
    while ns:
        if ns[i] == 0:
            n_metas = ns[i + 1]
            metas = ns[i + 2 : i + 2 + n_metas]
            sum_all_metas += sum(metas)
            del ns[i : i + 2 + n_metas]
            ns[i - 2] -= 1
            i -= 2
            node.metas = metas
            node = node.parent
        else:
            i += 2
            node = MyNode(parent=node)
except IndexError:
    assert node is root
    root.metas = metas
    print("answer_a:", sum_all_metas)
    print("answer_b:", node.val)



"""
--- Day 12: The N-Body Problem ---
https://adventofcode.com/2019/day/12
"""
import re
from itertools import count
from math import lcm

import numpy as np
from aocd import data
from aocd import extra


def simulate(data, n=0):
    p = np.array(re.findall(r"-?\d+", data)).astype(int).reshape(-1, 3)
    v = np.zeros_like(p)

    ts = [0, 0, 0]
    ss = [(p[:, i].tobytes(), v[:, i].tobytes()) for i in range(3)]
    for t in range(n) or count():

        for i in range(3):
            if not ts[i]:
                s = p[:, i].tobytes(), v[:, i].tobytes()
                if s == ss[i]:
                    ts[i] = t
                    if all(ts):
                        tx, ty, tz = ts
                        return lcm(tx, ty, tz)

        v += np.sign(p[:, None, :] - p).sum(axis=0)
        p += v

    pe = abs(p).sum(axis=1)
    ke = abs(v).sum(axis=1)
    e = (pe * ke).sum()
    return e


n = extra.get("iterations", 1000)
print("answer_a:", simulate(data, n))
print("answer_b:", simulate(data))



"""
--- Day 5: Cafeteria ---
https://adventofcode.com/2025/day/5
"""

from aocd import data

rs, ns = data.split("\n\n")
ns = [int(x) for x in ns.split()]
rs = sorted([[*map(int, x.split("-"))] for x in rs.split()])
merged = [rs[0]]
for r in rs:
    if r[0] > merged[-1][1]:
        merged.append(r)
    else:
        merged[-1][1] = max(merged[-1][1], r[1])

print("answer_a:", sum(any(x <= n <= y for x, y in merged) for n in ns))
print("answer_b:", sum(y - x + 1 for x, y in merged))



"""
--- Day 14: One-Time Pad ---
https://adventofcode.com/2016/day/14
"""
import os
import re
from collections import defaultdict
from collections import deque
from concurrent.futures import ProcessPoolExecutor
from itertools import count

from aocd import data

from aoc_wim.aoc2016 import md5_miner_q14a
from aoc_wim.aoc2016 import md5_miner_q14b


def search(data, stretch):
    keys = []
    triples = defaultdict(list)
    pat3 = re.compile(r"(.)\1{2}")
    pat5 = re.compile(r"(.)\1{4}")
    i = 0
    stop = None
    miner = md5_miner_q14b if stretch else md5_miner_q14a
    chunksize = 50 if stretch else 50000
    blocks = ((data, i0, chunksize) for i0 in count(step=chunksize))
    with ProcessPoolExecutor() as pool:
        futures = deque([pool.submit(miner, *next(blocks)) for _ in range(os.cpu_count())])
        while True:
            for i, md5sum in futures.popleft().result():
                triple = pat3.search(md5sum)
                for quintuple in pat5.findall(md5sum):
                    keys.extend([x for x in triples[quintuple] if i - x <= 1000])
                    triples[quintuple].clear()  # avoid to count same key twice
                    if stop is None and len(keys) >= 64:
                        stop = i + 1001
                triples[triple.group()[0]].append(i)
            if stop is not None and i > stop:
                break
            futures.append(pool.submit(miner, *next(blocks)))
        return sorted(keys)[63]


if __name__ == "__main__":
    print("answer_a:", search(data, stretch=False))
    print("answer_b:", search(data, stretch=True))



"""
--- Day 6: Guard Gallivant ---
https://adventofcode.com/2024/day/6
"""
from aocd import data

from aoc_wim.zgrid import ZGrid


def walk(z, dz):
    n = 0
    path = {(z, dz): n}
    cyclic = False
    while z + dz in grid:
        while grid[z + dz] == "#":
            dz *= 1j
        z += dz
        n += 1
        if (z, dz) in path:
            cyclic = True
            break
        path[z, dz] = n
    return path, cyclic


grid = ZGrid(data)
path, _cyclic = walk(grid.z("^"), -1j)
path = list(path)
print("answer_a:", len(dict(path)))

b = 0
seen = {path[0][0]}
for (z0, dz0), (z1, _) in zip(path, path[1:]):
    if z1 not in seen:
        grid[z1] = "#"
        b += walk(z0, dz0)[1]
        grid[z1] = "."
        seen.add(z1)
print("answer_b:", b)



"""
--- Day 4: Printing Department ---
https://adventofcode.com/2025/day/4
"""

from aocd import data
from aoc_wim.zgrid import ZGrid

grid = ZGrid(data)
zs = [z for z in grid.z("@", 0) if grid.count_near(z, "@", n=8) < 4]
a = b = len(zs)
print("answer_a:", a)

while zs:
    grid.update(dict.fromkeys(zs, "x"))
    zs = [z for z in grid.z("@", 0) if grid.count_near(z, "@", n=8) < 4]
    b += len(zs)
print("answer_b:", b)



"""
--- Day 13: Knights of the Dinner Table ---
https://adventofcode.com/2015/day/13
"""
from collections import defaultdict
from itertools import permutations

from aocd import data


def parsed(data, extra_name=None):
    d = defaultdict(int)
    names = {extra_name} - {None}
    for line in data.splitlines():
        words = line.split()
        name0 = words[0]
        name1 = words[-1].rstrip(".")
        n = {"gain": 1, "lose": -1}[words[2]] * int(words[3])
        d[(name0, name1)] = n
        names |= {name0, name1}
    return names, d


def get_best_plan(data, extra_name=None):
    names, d = parsed(data, extra_name)
    n = len(names)
    plans = permutations(names)
    happiness = {}
    for plan in plans:
        total = 0
        for i in range(n):
            person = plan[i]
            left = plan[(i - 1) % n]
            right = plan[(i + 1) % n]
            total += d[(person, left)]
            total += d[(person, right)]
        happiness[plan] = total
    return max(happiness.values())


print("answer_a:", get_best_plan(data))
print("answer_b:", get_best_plan(data, extra_name="wim"))



"""
--- Day 17: No Such Thing as Too Much ---
https://adventofcode.com/2015/day/17
"""
from aocd import data
from aocd import extra

from aoc_wim.stuff import subset_sum


vals = [*map(int, data.split())]
subsets = list(subset_sum(vals, extra.get("liters", 150)))
print("answer_a:", len(subsets))
fewest = min(subsets, key=len)
print("answer_b:", sum(len(s) == len(fewest) for s in subsets))



"""
--- Day 7: Bridge Repair ---
https://adventofcode.com/2024/day/7
"""
from math import log10

from aocd import data


def possible(target, ns, part="a"):
    *ns, n = ns
    if not ns:
        return n == target
    if part == "b":
        q, r = divmod(target, 10 ** (1 + int(log10(n))))
        if r == n and possible(q, ns, part):
            return True
    q, r = divmod(target, n)
    if r == 0 and possible(q, ns, part):
        return True
    return possible(target - n, ns, part)


a = b = 0
for line in data.replace(":", "").splitlines():
    target, *ns = map(int, line.split())
    if possible(target, ns):
        a += target
        b += target
    elif possible(target, ns, part="b"):
        b += target


print("answer_a:", a)
print("answer_b:", b)



"""
--- Day 8: I Heard You Like Registers ---
https://adventofcode.com/2017/day/8
"""
from aocd import data


class crazydict(dict):
    def __missing__(self, key):
        if key == "wtf":
            return locals
        if key == "max":
            return max
        self[key] = 0
        return 0


data += "\n"
data = data.replace("\n", " else 0\nΣ = max(Σ, *wtf().values())\n")
data = data.replace("inc", "+=").replace("dec", "-=")
d = crazydict()
exec(data, {}, d)
b = d.pop("Σ")
print("answer_a:", max(d.values()))
print("answer_b:", b)



"""
--- Day 21: Keypad Conundrum ---
https://adventofcode.com/2024/day/21
"""
from collections import Counter

import networkx as nx
from aocd import data

from aoc_wim.zgrid import ZGrid


keypad_numeric = """\
789
456
123
 0A
"""
keypad_directional = """\
 ^A
<v>
"""
dzs = dict(zip([1, 1j, -1, -1j], ">v<^"))
glyph0 = "0123456789A"
glyph1 = "<^v>A"
grid0 = ZGrid(keypad_numeric)
grid1 = ZGrid(keypad_directional)
graph0 = grid0.graph(extra=glyph0)
graph1 = grid1.graph(extra=glyph1)


def step_counts(z0, z1, graph):
    paths = []
    for path in nx.all_shortest_paths(graph, z0, z1):
        d_path = "".join(dzs[b - a] for a, b in zip(path, path[1:])) + "A"
        paths.append(d_path)
    seq = [sum(x == y for x, y in zip(p, p[1:])) for p in paths]
    max_seq = max(seq)
    paths = [p for s, p in zip(seq, paths) if s == max_seq]
    path = min(paths, key=lambda p: glyph1.index(p[0]))
    counts = Counter(c0 + c1 for c0, c1 in zip("A" + path, path))
    return counts


steps = {}
for src in glyph0:
    for dst in glyph0:
        steps[src + dst] = step_counts(grid0.z(src), grid0.z(dst), graph0)
for src in glyph1:
    for dst in glyph1:
        steps[src + dst] = step_counts(grid1.z(src), grid1.z(dst), graph1)


def complexity(code, n):
    tz0 = sum([steps[c0 + c1] for c0, c1 in zip("A" + code, code)], Counter())
    for i in range(n):
        tz = Counter()
        for k0, n0 in tz0.items():
            ck = steps[k0]
            tz += Counter({k1: n0 * n1 for k1, n1 in ck.items()})
        tz0 = tz
    path_length = sum(tz0.values())
    result = int(code[:-1]) * path_length
    return result


a = sum([complexity(code, n=2) for code in data.split()])
b = sum([complexity(code, n=25) for code in data.split()])
print("answer_a:", a)
print("answer_b:", b)



"""
--- Day 24: Planet of Discord ---
https://adventofcode.com/2019/day/24
"""
from aocd import data
from aocd import extra

from aoc_wim.zgrid import ZGrid


def new_empty_grid():
    grid = ZGrid(".....\n" * 5)
    del grid[2 + 2j]
    return grid


def evolve_a(grid):
    d = {}
    for z, g in grid.items():
        n = grid.count_near(z, val="#")
        if g == "#":
            d[z] = ".#"[n == 1]
        elif 1 <= n <= 2:
            d[z] = "#"
        else:
            d[z] = g
    grid.d = d


def part_a(data):
    seen = set()
    grid = ZGrid(data)
    while True:
        k = tuple(grid.items())
        if k in seen:
            return sum(2 ** i for i, glyph in enumerate(grid.values()) if glyph == "#")
        seen.add(k)
        evolve_a(grid)


def evolve_b(grids):
    min_grid = min(grids)
    max_grid = max(grids)
    grids[min_grid - 1] = new_empty_grid()  # add new grid above
    grids[max_grid + 1] = new_empty_grid()  # add new grid below
    new_grids = {}
    for depth, grid in grids.items():
        d = {}
        for z0, glyph in grid.items():
            n = 0
            for z in grid.near(z0):
                if z == 2 + 2j:
                    if depth + 1 not in grids:
                        continue
                    grid_down = grids[depth + 1]
                    if z == z0 + ZGrid.down:
                        n += sum([grid_down[i] == "#" for i in range(5)])
                    elif z == z0 + ZGrid.left:
                        n += sum([grid_down[4 + i * 1j] == "#" for i in range(5)])
                    elif z == z0 + ZGrid.up:
                        n += sum([grid_down[i + 4j] == "#" for i in range(5)])
                    elif z == z0 + ZGrid.right:
                        n += sum([grid_down[i * 1j] == "#" for i in range(5)])
                elif z in grid:
                    n += grid[z] == "#"
                else:
                    if depth - 1 not in grids:
                        continue
                    grid_up = grids[depth - 1]
                    assert z not in grid
                    if z.imag < 0:
                        n += grid_up[2 + 1j] == "#"
                    if z.imag > 4:
                        n += grid_up[2 + 3j] == "#"
                    if z.real < 0:
                        n += grid_up[1 + 2j] == "#"
                    if z.real > 4:
                        n += grid_up[3 + 2j] == "#"
            if glyph == "#":
                if n == 1:
                    d[z0] = "#"
                else:
                    d[z0] = "."
            elif 1 <= n <= 2:
                d[z0] = "#"
            else:
                d[z0] = glyph
        new_grids[depth] = ZGrid(d)
    if "#" not in new_grids[min_grid - 1].values():
        del new_grids[min_grid - 1]
    if "#" not in new_grids[max_grid + 1].values():
        del new_grids[max_grid + 1]
    grids.clear()
    grids.update(new_grids)


def part_b(data):
    zgrid = ZGrid(data)
    t = extra.get("iterations", 200)
    del zgrid[2 + 2j]
    grids = {0: zgrid}
    for i in range(t):
        evolve_b(grids)
    n_bugs = 0
    for depth, grid in sorted(grids.items()):
        # print(f"Depth {depth}:")
        # grid.draw(pretty=False, overlay={2+2j: "?"})
        n_bugs += grid.count("#")
    return n_bugs


if __name__ == "__main__":
    print("answer_a:", part_a(data))
    print("answer_b:", part_b(data))



"""
--- Day 15: Dueling Generators ---
https://adventofcode.com/2017/day/15
"""
import numpy as np
from aocd import data
from parse import parse

template = (
    "Generator A starts with {:d}\n"
    "Generator B starts with {:d}"
)
a0, b0 = parse(template, data).fixed
fa = 16807
fb = 48271
d = 0x7fffffff

AB = np.array([[a0, b0]], dtype=np.int64) * (fa, fb) % d
while len(AB) < 40_000_000:
    f = pow(fa, len(AB), d), pow(fb, len(AB), d)
    AB = np.vstack([AB, AB * f % d])

AB &= 0xffff
A, B = AB.T
A4 = A[A & 0b11 == 0]
B8 = B[B & 0b111 == 0]
print("answer_a:", (A[:40_000_000] == B[:40_000_000]).sum())
print("answer_b:", (A4[:5_000_000] == B8[:5_000_000]).sum())



"""
--- Day 18: RAM Run ---
https://adventofcode.com/2024/day/18
"""
import networkx as nx
from aocd import data
from aocd import extra

from aoc_wim.zgrid import ZGrid


w = extra.get("width", 71)
n_bytes = extra.get("n_bytes", 1024)
grid = ZGrid.fromempty(w, w)
zs = [complex(*map(int, x.split(","))) for x in data.split()]

grid.update({z: "#" for z in zs[:n_bytes]})
graph = grid.graph()
path = nx.shortest_path(graph, 0, complex(w-1, w-1))
print("answer_a:", len(path) - 1)

for z in zs[n_bytes:]:
    grid[z] = "#"
    if z not in path:
        continue
    graph = grid.graph()
    try:
        path = nx.shortest_path(graph, 0, complex(w-1, w-1))
    except nx.NetworkXNoPath:
        print("answer_b:", f"{int(z.real)},{int(z.imag)}")
        break



"""
--- Day 19: Monster Messages ---
https://adventofcode.com/2020/day/19
"""
from aocd import data


def match(message, remaining):
    if not message or not remaining:
        return not message and not remaining
    first, *rest = remaining
    rule = rules[first]
    if rule in ("a", "b"):
        if message.startswith(rule):
            return match(message[1:], rest)
        return False
    return any(match(message, t + rest) for t in rule)


rules_raw, messages = data.replace('"', "").split("\n\n")
messages = messages.splitlines()
rules = dict(r.split(": ") for r in rules_raw.splitlines())
for k, v in rules.items():
    if v not in "ab":
        rules[k] = [t.split() for t in v.split("|")]

print("answer_a:", sum(match(m, ["0"]) for m in messages))

rules["8"].append(["42", "8"])
rules["11"].append(["42", "11", "31"])
print("answer_b:", sum(match(m, ["0"]) for m in messages))



"""Wim's solutions for https://adventofcode.com/"""
import json
import logging
import os
import runpy
import subprocess
import sys
from argparse import ArgumentParser
from datetime import datetime
from pathlib import Path

import aocd
from aocd.models import Puzzle
from aocd.utils import AOC_TZ

from aoc_wim.util import parse_extra_context
from aoc_wim.util import split_trailing_comments


here = Path(__file__).parent


def run_one():
    parser = ArgumentParser(description=__doc__)
    aoc_now = datetime.now(tz=AOC_TZ)
    days = range(1, 26)
    years = range(2015, aoc_now.year + int(aoc_now.month == 12))
    parser.add_argument(
        "day",
        nargs="?",
        type=int,
        default=min(aoc_now.day, 25) if aoc_now.month == 12 else 1,
        help="1-25 (default: %(default)s)",
    )
    parser.add_argument(
        "year",
        nargs="?",
        type=int,
        default=years[-1],
        help="2015-%(default)s (default: %(default)s)",
    )
    parser.add_argument("-u", "--user", choices=["github", "twitter", "reddit", "google"])
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-d", "--data", help="string or file to monkeypatch in aocd.data")
    group.add_argument("-t", "--test", help="use the example data (if any)", action="store_true")
    log_levels = "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"
    parser.add_argument("--log-level", type=str.upper, choices=log_levels)
    args = parser.parse_args()
    if args.day in years and args.year in days:
        args.day, args.year = args.year, args.day
    if args.log_level:
        level_int = getattr(logging, args.log_level)
        logging.basicConfig(format="%(message)s", level=level_int)
    mod_name = f"aoc_wim.aoc{args.year}.q{args.day:02d}"
    sys.modules.pop(mod_name, None)
    if args.user is not None:
        users = json.loads(Path("~/.config/aocd/tokens.json").expanduser().read_text())
        [token] = [v for k, v in users.items() if k.startswith(args.user)]
        os.environ["AOC_SESSION"] = token
    if args.data is not None:
        if args.data.endswith("txt"):
            path = Path(args.data)
            if path.is_file():
                txt = path.read_text()
                if "tests" in path.parts:
                    args.data = "\n".join(txt.splitlines()[:-2])
                else:
                    args.data = txt
        aocd.data = args.data
    elif args.test:
        test_dir = here.parent / "tests" / f"{args.year}" / f"{args.day:02d}"
        test_files = list(test_dir.glob("*.txt"))
        if not test_files:
            sys.exit(f"no test files found at {test_dir}")
        for test_path in sorted(test_files):
            print(f"\n(using data from {test_path})")
            print(f"--- {args.year} Day {args.day}: {Puzzle(args.year, args.day).title} ---")
            lines = test_path.read_text().splitlines()
            aocd.data = "\n".join(lines[:-2])
            aocd.extra = parse_extra_context(split_trailing_comments(lines))
            runpy.run_module(mod_name, run_name="__main__")
        print()
        return
    puzzle = Puzzle(args.year, args.day)
    print(f"--- {args.year} Day {args.day}: {puzzle.title} ---")
    runpy.run_module(mod_name, run_name="__main__")


def speedhack():
    here = Path(__file__).parent
    aoc_now = datetime.now(tz=AOC_TZ)
    if aoc_now.month != 12:
        sys.exit("It's not December yet")
    year = aoc_now.year
    day = aoc_now.day
    test_dir = here.parent / "tests" / str(year) / str(day).zfill(2)
    examples = sorted(test_dir.glob("*.txt"))
    puzzle = Puzzle(year, day)
    print(f"--- {year} Day {day}: {puzzle.title} ---")
    print(puzzle.url)
    if examples:
        print(f"{len(examples)} example(s) to test...")
        args = [sys.executable, "-m", "pytest", "-k", f"{year}/{day:02d}"]
        subprocess.check_call(args, cwd=Path("~/git/advent-of-code-wim").expanduser())
    args = ["aoc", "-y", str(year), "-d", str(day), "-u", "github.wimglenn.119932", "-r", "-vv", "--timeout", "6000"]
    print("\n" + " ".join(args))
    subprocess.check_call(args)



"""
--- Day 22: Monkey Map ---
https://adventofcode.com/2022/day/22
"""
import math
import re

from aocd import data

from aoc_wim.zgrid import ZGrid


grid, steps = data.split("\n\n")
grid = ZGrid(grid)
steps = re.split("([LR])", steps)

grid.remove(" ")
dzs = [1, 1j, -1, -1j]
glyph = dict(zip(dzs, ">v<^"))
w = math.isqrt(len(grid) // 6)


def passwd(z, dz):
    return int(1000*(z.imag + 1) + 4*(z.real + 1) + dzs.index(dz))


def wrap_a(z, dz):
    z += dz
    while z - dz in grid:
        z -= w * dz
    return z, dz


def wrap_b_test(z, dz):
    """
            1111
            1111
            1111
            1111
    222233334444
    222233334444
    222233334444
    222233334444
            55556666
            55556666
            55556666
            55556666
    """
    X, x = divmod(z.real, w)
    Y, y = divmod(z.imag, w)
    match dz:
        case 1: return z + (1+1j)*w - y*(1+1j), 1j         # 4 -> 6
        case 1j: return z - (2+1j)*w + w-1-2*x, -1j        # 5 -> 2
        case -1j: return z - (1j-1)*w - (w-1-x)*(1-1j), 1  # 3 -> 1


def wrap_b(z, dz):
    """
        11112222
        11112222
        11112222
        11112222
        3333
        3333
        3333
        3333
    44445555
    44445555
    44445555
    44445555
    6666
    6666
    6666
    6666
    """
    X, x = divmod(z.real, w)
    Y, y = divmod(z.imag, w)
    x_ = w - 1 - x
    y_ = w - 1 - y
    match X, Y, dz:
        case _, 0,  -1: return z - w*(1-2j) + (y_-y)*1j,   1  # 1 -> 4
        case _, 1,  -1: return z - w*(1-1j) + y*(1-1j),   1j  # 3 -> 4
        case _, 2,  -1: return z + w*(1-2j) + (y_-y)*1j,   1  # 4 -> 1
        case _, 3,  -1: return z + w*(1-3j) + y*(1-1j),   1j  # 6 -> 1
        case _, 0,   1: return z - w*(1-2j) + (y_-y)*1j,  -1  # 2 -> 5
        case _, 1,   1: return z + w*(1-1j) - y_*(1-1j), -1j  # 3 -> 2
        case _, 2,   1: return z + w*(1-2j) + (y_-y)*1j,  -1  # 5 -> 2
        case _, 3,   1: return z + w*(1-1j) - y_*(1-1j), -1j  # 6 -> 5
        case 0, _, -1j: return z + w*(1-1j) - x*(1-1j),    1  # 4 -> 3
        case 1, _, -1j: return z - w*(1-3j) - x*(1-1j),    1  # 1 -> 6
        case 2, _, -1j: return z - w*(2-3j) + (w-1)*1j,  -1j  # 2 -> 6
        case 0, _,  1j: return z + w*(2-3j) - (w-1)*1j,   1j  # 6 -> 2
        case 1, _,  1j: return z - w*(1-1j) + x_*(1-1j),  -1  # 5 -> 6
        case 2, _,  1j: return z - w*(1-1j) + x_*(1-1j),  -1  # 2 -> 3


def walk_path(z, dz, wrap, draw=False):
    path_overlay = {z: glyph[dz]}
    for step in steps:
        if step == "R":
            dz *= 1j
        elif step == "L":
            dz *= -1j
        else:
            for _ in range(int(step)):
                z_ = z + dz
                dz_ = dz
                if z_ not in grid:
                    z_, dz_ = wrap(z, dz)
                if grid[z_] == "#":
                    break
                z = z_
                dz = dz_
                path_overlay[z] = glyph[dz]
        path_overlay[z] = glyph[dz]
    if draw:
        grid.draw(overlay=path_overlay, empty_glyph=" ")
    return z, dz


z0 = grid.z(".")
dz0 = 1

draw = False
if w == 4:
    wrap_b = wrap_b_test
    draw = True

z, dz = walk_path(z0, dz0, wrap=wrap_a, draw=draw)
print("answer_a:", passwd(z, dz))

z, dz = walk_path(z0, dz0, wrap=wrap_b, draw=draw)
print("answer_b:", passwd(z, dz))



"""
--- Day 13: Claw Contraption ---
https://adventofcode.com/2024/day/13
"""
import parse
from aocd import data

template = """\
Button A: X+{:d}, Y+{:d}
Button B: X+{:d}, Y+{:d}
Prize: X={:d}, Y={:d}"""
ab = [0, 0]
for game in parse.findall(template, data):
    ax, ay, bx, by, X, Y = game.fixed
    denom = ax * by - bx * ay
    for o in 0, 10**13:
        pa, ra = divmod(by * (X + o) - bx * (Y + o), denom)
        pb, rb = divmod(ax * (Y + o) - ay * (X + o), denom)
        if not ra and not rb:
            ab[o > 0] += 3 * pa + pb
a, b = ab

print("answer_a:", a)
print("answer_b:", b)



"""
--- Day 6: Trash Compactor ---
https://adventofcode.com/2025/day/6
"""

from math import prod
from aocd import data

lines = data.splitlines()
d = [x.split() for x in lines]
dT = [list(x) for x in zip(*d)]
a = 0
for col in dT:
    op = prod if col.pop() == "*" else sum
    a += op([int(x) for x in col])
print("answer_a:", a)

ops = lines.pop().split()
lines = [" " + x for x in lines]
[L] = {len(x) for x in lines}
ns = []
b = 0
for i in reversed(range(L)):
    s = "".join([x[i] for x in lines]).strip()
    if s:
        ns.append(int(s))
    else:
        op = prod if ops.pop() == "*" else sum
        b += op(ns)
        ns.clear()
print("answer_b:", b)



"""
--- Day 7: Some Assembly Required ---
https://adventofcode.com/2015/day/7
"""
import re
from collections import deque

import numpy as np
from aocd import data


def compute(data):
    wire = {}

    def get(v):
        return wire[v] if v in wire else np.uint16(v)

    q = deque(data.splitlines())
    while q:
        line = q.pop()
        try:
            match line.split():
                case n, "->", w:
                    wire[w] = get(n)
                case "NOT", n, "->", w:
                    wire[w] = ~get(n)
                case x, op, y, "->", w:
                    f = getattr(np.uint16, f"__{op}__".lower())
                    wire[w] = f(get(x), get(y))
        except (KeyError, ValueError):
            q.appendleft(line)

    return wire["a"]


a = compute(data)
print("answer_a:", a)

new_data = re.sub(r"\n([0-9]+) -> b\n", f"\n{a} -> b\n", data)
print("answer_b:", compute(new_data))



"""
--- Day 25: Snowverload ---
https://adventofcode.com/2023/day/25
"""
import networkx as nx
from aocd import data


graph = nx.Graph()
for line in data.replace(":", "").splitlines():
    left, *rights = line.split()
    for right in rights:
        graph.add_edge(left, right)

ebc = nx.edge_betweenness_centrality(graph)
for i in range(3):
    edge = max(ebc, key=ebc.get)
    graph.remove_edge(*edge)
    del ebc[edge]

comp1, comp2 = nx.connected_components(graph)
print("answer_a:", len(comp1) * len(comp2))



"""
--- Day 4: The Ideal Stocking Stuffer ---
https://adventofcode.com/2015/day/4
"""
from _md5 import md5

from aocd import data


def mine(data, difficulty, start=0):
    secret_key = data.encode()
    prefix = "0" * difficulty
    h0 = md5(secret_key).copy
    i = start
    while True:
        i += 1
        hash_ = h0()
        hash_.update(b"%d" % i)
        if hash_.hexdigest().startswith(prefix):
            return i


a = mine(data, difficulty=5)
print("answer_a:", a)
b = mine(data, difficulty=6, start=a - 1)
print("answer_b:", b)



"""
--- Day 22: Sporifica Virus ---
https://adventofcode.com/2017/day/22
"""
from aocd import data
from aocd import extra


def parsed(data):
    lines = data.splitlines()
    s = len(lines) // 2
    grid = {}
    for row, line in enumerate(lines, -s):
        for col, char in enumerate(line, -s):
            grid[col - row * 1j] = char
    return grid


def mutate(data, n, part="a"):
    mutations = {
        "a": {"#": ".", ".": "#"},
        "b": {"#": "F", "F": ".", ".": "W", "W": "#"},
    }
    mutation = mutations[part]
    factors = {".": 1j, "W": 1, "F": -1, "#": -1j}
    grid = parsed(data)
    p, v = 0, 1j
    n_infected = 0
    for i in range(n):
        s = grid.get(p, ".")
        v *= factors[s]
        grid[p] = mutation[s]
        if grid[p] == "#":
            n_infected += 1
        p += v
    return n_infected


if __name__ == "__main__":
    print("answer_a:", mutate(data, n=extra.get("iterations", 10000), part="a"))
    print("answer_b:", mutate(data, n=extra.get("iterations", 10000000), part="b"))



"""
--- Day 18: Like a Rogue ---
https://adventofcode.com/2016/day/18
"""
from aocd import data
from aocd import extra


def generate_rows(row):
    """rule90: https://en.wikipedia.org/wiki/Rule_90"""
    idx = range(len(row))
    while True:
        yield row
        row = "." + row + "."
        row = "".join(["^."[row[i] == row[i + 2]] for i in idx])


n_rows = extra.get("n_rows", 40)
gen = generate_rows(data)
safe = sum(next(gen).count(".") for _ in range(n_rows))
print("answer_a:", safe)
safe += sum(next(gen).count(".") for _ in range(400000 - n_rows))
print("answer_b:", safe)



"""
--- Day 23: Amphipod ---
https://adventofcode.com/2021/day/23
"""
import networkx as nx
from aocd import data

from aoc_wim.search import AStar
from aoc_wim.zgrid import manhattan_distance
from aoc_wim.zgrid import ZGrid


class Q23AStar(AStar):
    def __init__(self, data):
        grid = ZGrid(data)
        graph = grid.graph(extra="ABCD")
        hall = self.hall = [z for z in graph if z.imag == 1]
        rooms = self.rooms = [z for z in graph if z.imag != 1]
        # Amphipods will never stop on the space immediately outside any room
        hall[:] = [z for z in hall if z + ZGrid.down not in rooms]
        obstructions = self.obstructions = {}
        for r in rooms:
            for h in hall:
                path = nx.shortest_path(graph, r, h)
                assert path[0] == r
                assert path[-1] == h
                assert len(path) == manhattan_distance(r, h) + 1
                obs = [z for z in path[1:-1] if z in rooms + hall]
                obstructions[r, h] = obstructions[h, r] = frozenset(obs)

        state0 = {c: [] for c in "ABCD"}
        for z, glyph in grid.items():
            if glyph in "ABCD":
                state0[glyph].append(z)
        state0 = tuple(frozenset(v) for v in state0.values())
        target = tuple(frozenset(rooms[i::4]) for i in range(4))
        AStar.__init__(self, state0, target)

    def adjacent(self, state):
        occupied = frozenset.union(*state)
        for i, ps in enumerate(state):
            target_rooms = self.rooms[i::4]
            for p in ps:
                if p.imag != 1:
                    # amphipod is in a side room
                    if p in target_rooms:
                        t = target_rooms.index(p)
                        others = frozenset(target_rooms[t:])
                        if others <= ps:
                            # this amphipod is already home - it shouldn't move again
                            continue
                    # it can move into a hallway position if not obstructed
                    for h in self.hall:
                        if h not in occupied:
                            if not self.obstructions[h, p].intersection(occupied):
                                new_ps = ps - {p} | {h}
                                yield *state[:i], new_ps, *state[i + 1 :]
                else:
                    # amphipod is in hallway
                    # it can move into target room if unobstructed
                    for j, t in enumerate(target_rooms):
                        if t not in occupied:
                            if ps.issuperset(target_rooms[j + 1 :]):
                                if not self.obstructions[p, t].intersection(occupied):
                                    new_ps = ps - {p} | {t}
                                    yield *state[:i], new_ps, *state[i + 1 :]
                                    break

    def cost(self, current_state, next_state):
        # find who has moved
        for g, s0, s1 in zip("ABCD", current_state, next_state):
            diff = s0 ^ s1
            if diff:
                return manhattan_distance(*diff) * 10 ** "ABCD".index(g)
        return 0


def render(state):
    g = ZGrid(data, transform=dict.fromkeys("ABCD", "."))
    overlay = {z: g for g, zs in zip("ABCD", state) for z in zs}
    g.draw(overlay=overlay)


def total_energy(data):
    astar = Q23AStar(data)
    astar.run()
    return astar.gscore[astar.target]


print("answer_a:", total_energy(data))
extra = """\
  #D#C#B#A#
  #D#B#A#C#
"""
lines = data.splitlines()
data = "\n".join(lines[:-2] + extra.splitlines() + lines[-2:])
print("answer_b:", total_energy(data))



"""
--- Day 8: Two-Factor Authentication ---
https://adventofcode.com/2016/day/8
"""
import numpy as np
from aocd import data
from aocd import extra

from aoc_wim.ocr import AOCR


def animate(A, line):
    if line.startswith("rect"):
        w, h = [int(x) for x in line.split()[1].split("x")]
        A[0:h, 0:w] = "#"
    elif line.startswith("rotate"):
        i, shift = [int(x) for x in line.split("=")[1].split(" by ")]
        item = (i, slice(None)) if "row" in line else (slice(None), i)
        A[item] = np.roll(A[item], shift)
    return A


lines = data.splitlines()
w = extra.get("screen_width", 50)
h = extra.get("screen_height", 6)
A = np.full((h, w), ".")
for line in lines:
    A = animate(A, line)
print("answer_a:", (A == "#").sum())
print("answer_b:", AOCR[A])



"""
--- Day 19: Linen Layout ---
https://adventofcode.com/2024/day/19
"""
from functools import cache

from aocd import data


@cache
def ways(design):
    if not design:
        return 1
    return sum(ways(design[len(p):]) for p in patterns if design.startswith(p))


patterns, designs = data.split("\n\n")
patterns = patterns.split(", ")
results = [ways(d) for d in designs.split()]
print("answer_a:", sum(map(bool, results)))
print("answer_b:", sum(results))



"""
--- Day 18: Like a GIF For Your Yard ---
https://adventofcode.com/2015/day/18
"""
import numpy as np
from aocd import data
from aocd import extra
from scipy.signal import convolve2d

from aoc_wim.zgrid import ZGrid


kernel = np.ones((3, 3), dtype=int)
kernel[1, 1] = 0


def evolve(A):
    """conway's game of life"""
    C = convolve2d(A, kernel, mode="same")
    A = ((A == 1) & ((C == 2) | (C == 3))) | ((A == 0) & (C == 3))
    return A.astype(int)


def animate(data, part="a"):
    grid = ZGrid(data, transform={"#": 1, ".": 0})
    A = np.array(grid)
    if part == "b":
        A[0, 0] = A[0, -1] = A[-1, 0] = A[-1, -1] = 1
    for i in range(extra.get("iterations", 100)):
        A = evolve(A)
        if part == "b":
            A[0, 0] = A[0, -1] = A[-1, 0] = A[-1, -1] = 1
    return A.sum()


print("answer_a:", animate(data, part="a"))
print("answer_b:", animate(data, part="b"))



"""
--- Day 4: Repose Record ---
https://adventofcode.com/2018/day/4
"""
from collections import Counter
from datetime import timedelta
from operator import attrgetter

from aocd import data
from parse import parse


class Guard:

    template = "[{t:ti}] Guard #{id:d} begins shift"
    eventstr = "[{t:ti}] {event}"

    def __init__(self, id, began_shift):
        self.id = id
        self.began_shift = began_shift
        self.t0 = None
        self.minutes_asleep = []

    @classmethod
    def fromline(cls, line):
        parsed = parse(cls.template, line)
        return Guard(id=parsed["id"], began_shift=parsed["t"])

    def event(self, line):
        parsed = parse(self.eventstr, line)
        t = parsed["t"]
        if parsed["event"] == "falls asleep":
            self.t0 = t
        elif parsed["event"] == "wakes up":
            t0 = self.t0
            while t0 < t:
                self.minutes_asleep.append(t0.minute)
                t0 += timedelta(minutes=1)
            self.t0 = None

    @property
    def total_sleep_time(self):
        return len(self.minutes_asleep)

    @property
    def sleepiest_minute(self):
        counter = Counter(self.minutes_asleep)
        return max(counter, key=counter.get)

    @property
    def sleepiest_minute_count(self):
        return max(Counter(self.minutes_asleep).values(), default=0)


guards = {}
for line in sorted(data.splitlines()):
    if "Guard" in line:
        guard = Guard.fromline(line)
        guard = guards.get(guard.id) or guard
        guards[guard.id] = guard
        continue
    guard.event(line)

g = max(guards.values(), key=attrgetter("total_sleep_time"))
print("answer_a:", g.id * g.sleepiest_minute)

g = max(guards.values(), key=attrgetter("sleepiest_minute_count"))
print("answer_b:", g.id * g.sleepiest_minute)



"""
--- Day 10: Elves Look, Elves Say ---
https://adventofcode.com/2015/day/10
"""
from itertools import groupby

from aocd import data
from aocd import extra


def look_and_say(s, n=1):
    for i in range(n):
        s = "".join([f"{len(list(group))}{key}" for key, group in groupby(s)])
    return s


n = extra.get("iterations", 40)
a = look_and_say(data, n=n)
print("answer_a:", len(a))
if not extra:
    b = look_and_say(a, n=50 - n)
    print("answer_b:", len(b))



"""
--- Day 16: Chronal Classification ---
https://adventofcode.com/2018/day/16
"""
from aocd import data
from parse import parse


def all_ops():
    def addr(R, a, b, c):
        R[c] = R[a] + R[b]

    def addi(R, a, b, c):
        R[c] = R[a] + b

    def mulr(R, a, b, c):
        R[c] = R[a] * R[b]

    def muli(R, a, b, c):
        R[c] = R[a] * b

    def banr(R, a, b, c):
        R[c] = R[a] & R[b]

    def bani(R, a, b, c):
        R[c] = R[a] & b

    def borr(R, a, b, c):
        R[c] = R[a] | R[b]

    def bori(R, a, b, c):
        R[c] = R[a] | b

    def setr(R, a, b, c):
        R[c] = R[a]

    def seti(R, a, b, c):
        R[c] = a

    def gtir(R, a, b, c):
        R[c] = int(a > R[b])

    def gtri(R, a, b, c):
        R[c] = int(R[a] > b)

    def gtrr(R, a, b, c):
        R[c] = int(R[a] > R[b])

    def eqir(R, a, b, c):
        R[c] = int(a == R[b])

    def eqri(R, a, b, c):
        R[c] = int(R[a] == b)

    def eqrr(R, a, b, c):
        R[c] = int(R[a] == R[b])

    return locals()


def choices(s):
    ops = all_ops()
    parsed = parse(template, s)
    R0 = list(parsed.fixed[0:4])
    instruction = parsed.fixed[4:8]
    R1 = list(parsed.fixed[8:12])
    choices = []
    for opname, f in ops.items():
        opcode, a, b, c = instruction
        R = R0[:]
        f(R, a, b, c)
        if R == R1:
            choices.append(opname)
    return choices


template = """\
Before: [{:d}, {:d}, {:d}, {:d}]
{:d} {:d} {:d} {:d}
After:  [{:d}, {:d}, {:d}, {:d}]"""


def main():
    ops = all_ops()
    samples, _sep, prog = data.partition("\n" * 4)
    samples = samples.split("\n" * 2)
    print("answer_a:", sum(len(choices(s)) >= 3 for s in samples))
    if not prog:
        return
    identified = {}
    while len(identified) < len(ops):
        for s in samples:
            cs = set(choices(s)) - identified.keys()
            if len(cs) == 1:
                [opname] = cs
                opcode = parse(template, s).fixed[4]
                identified[opname] = opcode
    R = [0, 0, 0, 0]
    ops = {opnum: ops[opname] for opname, opnum in identified.items()}
    for line in prog.splitlines():
        opnum, a, b, c = [int(x) for x in line.split()]
        op = ops[opnum]
        op(R, a, b, c)
    print("answer_b:", R[0])


if __name__ == "__main__":
    main()



"""
--- Day 21: Step Counter ---
https://adventofcode.com/2023/day/21
"""
from aocd import data
from aocd import extra

from aoc_wim.search import BFS
from aoc_wim.zgrid import manhattan_distance
from aoc_wim.zgrid import ZGrid


grid = ZGrid(data)
z0 = grid.z("S")
grid[z0] = "."
plots = set(grid.bfs(z0=z0))


def step(n):
    parity = n % 2
    reachable = far = 0
    for z in plots:
        d = manhattan_distance(z, z0)
        if d > n:
            far += 1
        else:
            reachable += d % 2 == parity
    return reachable, far


def adj(state, h=grid.height, w=grid.width):
    z0, Z0 = state
    for z in grid.near(z0):
        if z in grid:
            Z = Z0
        else:
            qy, ry = divmod(z.imag, h)
            qx, rx = divmod(z.real, w)
            z = complex(rx, ry)
            Z = Z0 + complex(qx, qy)
        if (z, Z) not in bfs.seen and grid[z] == ".":
            yield z, Z


if n_steps := extra.get("n_steps"):
    state0 = z0, 0
    bfs = BFS(adj,  max_depth=n_steps)
    bfs(state0)
    parity = n_steps % 2
    a = b = sum(v % 2 == parity for v in bfs.seen.values())
    print("answer_a:", a)
    print("answer_b:", b)
else:
    a, _ = step(64)
    print("answer_a:", a)
    N, rem = divmod(26501365, grid.width)
    r, f = step(rem)
    b = N * N * len(plots) + (2 * N + 1) * r + N * f
    print("answer_b:", b)



"""
--- Day 11: Plutonian Pebbles ---
https://adventofcode.com/2024/day/11
"""
from collections import Counter

from aocd import data


def blink(d):
    result = Counter()
    for k, n in d.items():
        if k == "0":
            result["1"] += n
        elif len(k) % 2 == 0:
            i = len(k) // 2
            result[k[:i].lstrip("0") or "0"] += n
            result[k[i:].lstrip("0") or "0"] += n
        else:
            result[str(int(k) * 2024)] += n
    return result


stones = Counter(data.split())
for i in range(25):
    stones = blink(stones)
print("answer_a:", sum(stones.values()))

for i in range(50):
    stones = blink(stones)
print("answer_b:", sum(stones.values()))



"""
--- Day 23: LAN Party ---
https://adventofcode.com/2024/day/23
"""
import itertools as it

import networkx as nx
from aocd import data


graph = nx.Graph(x.split("-") for x in data.split())
triples = set()
for t in graph:
    if t.startswith("t"):
        for u, v in it.combinations(graph[t], 2):
            if (u, v) in graph.edges():
                triples.add(frozenset([t, u, v]))
print("answer_a:", len(triples))

clique = max(nx.find_cliques(graph), key=len)
print("answer_b:", ",".join(sorted(clique)))



"""
--- Day 8: Space Image Format ---
https://adventofcode.com/2019/day/8
"""
import numpy as np
from aocd import data
from aocd import extra

from aoc_wim.ocr import AOCR

w = extra.get("image_width", 25)
h = extra.get("image_height", 6)
a = np.fromiter(data, int).reshape((-1, h, w))
layer = min(a, key=lambda v: (v == 0).sum())
print("answer_a:", (layer == 1).sum() * (layer == 2).sum())

img = np.ones_like(layer) * 2
for layer in a:
    np.copyto(img, layer, where=(img == 2))

if w != 3:
    print("answer_b:", AOCR[img])



"""
--- Day 16: Permutation Promenade ---
https://adventofcode.com/2017/day/16
"""
from aocd import data
from aocd import extra


def parsed(data, mod=16):
    S = 0
    ops = []
    for op in data.split(","):
        if op.startswith("s"):
            n = int(op[1:])
            S += n
        elif op.startswith("x"):
            a, b = op[1:].split("/")
            a = (int(a) - S) % mod
            b = (int(b) - S) % mod
            ops.append(("x", a, b))
        elif op.startswith("p"):
            a, b = op[1:].split("/")
            ops.append(("p", a, b))
    ops.append(("s", S % mod, S % mod))
    return ops


def dance(data, d, n=1):
    mod = len(d)
    ops = parsed(data, mod=mod)
    seen = {}
    while n:
        s = "".join(d)
        if s in seen:
            r = seen[s] - n
            n %= r
            seen.clear()
            continue
        seen[s] = n
        for i, a, b in ops:
            if i == "x":
                d[a], d[b] = d[b], d[a]
            elif i == "p":
                ai = d.index(a)
                bi = d.index(b)
                d[ai], d[bi] = d[bi], d[ai]
            elif i == "s":
                d[:] = d[(mod - a) :] + d[: (mod - a)]
        n -= 1
    return "".join(d)


iterations = extra.get("iterations", 1000000000)
s0 = "abcdefghijklmnop"[:extra.get("n_programs")]

print("answer_a:", dance(data, list(s0)))
print("answer_b:", dance(data, list(s0), n=iterations))



"""
--- Day 25: Code Chronicle ---
https://adventofcode.com/2024/day/25
"""
from aocd import data

from aoc_wim.zgrid import ZGrid


chunks = data.split("\n\n")
keys = []
locks = []
for chunk in chunks:
    zs = set(ZGrid(chunk).z("#", first=False))
    L = keys if chunk[0] == "." else locks
    L.append(zs)

a = sum(1 for lock in locks for key in keys if not lock & key)
print("answer_a:", a)



"""
--- Day 16: Flawed Frequency Transmission ---
https://adventofcode.com/2019/day/16
"""
from itertools import cycle

import numpy as np
from aocd import data
from aocd import extra


def square_wave(w=1):
    base = [0] * w + [1] * w + [0] * w + [-1] * w
    gen = cycle(base)
    next(gen)
    return gen


def part_a(data):
    x = np.fromiter(data, int)
    [w] = x.shape
    waves = [square_wave(i + 1) for i in range(w)]
    A = np.array([[next(g) for _ in range(w)] for g in waves], dtype=int)
    for _ in range(extra.get("iterations", 100)):
        x = abs((A * x).sum(axis=1)) % 10
    result = ''.join(str(n) for n in x[:8])
    return result


def part_b(data):
    x = np.fromiter(data, int)
    [w] = x.shape
    pos = int(data[:7])
    pos_r = pos - w * 10_000
    stack = -pos_r // w + 1
    if stack <= 0:
        return
    X = np.hstack([x] * stack)
    X = X[pos_r:]
    for _ in range(extra.get("iterations", 100)):
        X = X[::-1].cumsum()[::-1] % 10
    return ''.join(str(n) for n in list(X)[pos_r:pos_r+8])


if __name__ == "__main__":
    print("answer_a:", part_a(data))
    print("answer_b:", part_b(data))



"""
--- Day 12: Digital Plumber ---
https://adventofcode.com/2017/day/12
"""
import networkx as nx
from aocd import data


graph = nx.Graph()
for line in data.replace(" <->", ",").splitlines():
    n0, *nodes = [int(n) for n in line.split(", ")]
    graph.add_node(n0)
    for node in nodes:
        graph.add_edge(n0, node)

[a] = [len(g) for g in nx.connected_components(graph) if 0 in g]
b = nx.number_connected_components(graph)

print("answer_a:", a)
print("answer_b:", b)



"""
--- Day 24: Crossed Wires ---
https://adventofcode.com/2024/day/24
"""
from collections import deque

from aocd import data
from aocd import extra


def parse(data):
    wires = {}
    ops = []
    for line in data.splitlines():
        match line.split():
            case wire, val:
                wires[wire.rstrip(":")] = int(val)
            case i0, op, i1, _, out:
                i0, i1 = sorted([i0, i1])
                ops.append([i0, op, i1, out])
    return wires, ops


def get_val(wires, name="z"):
    return sum(bit << int(w[1:]) for w, bit in wires.items() if w[0] == name)


def compute(wires, ops):
    q = deque(ops)
    while q:
        L = i0, op, i1, out = q.popleft()
        if i0 not in wires or i1 not in wires:
            q.append(L)  # procrastinate
            continue
        x, y = wires[i0], wires[i1]
        wires[out] = x & y if op == "AND" else x | y if op == "OR" else x ^ y
    return get_val(wires)


wires, ops = parse(data)
a = compute(wires, ops)
print("answer_a:", a)

swapped = []
n_swapped = 2 * extra.get("n_swapped_pairs", 4)
zs = sorted([w for w in wires if w[0] == "z"])
if extra.get("operation") == "bitwise_and":
    d0 = {}
    for z in zs:
        d0[z.replace("z", "x")] = d0[z.replace("z", "y")] = 0
    for z in zs:
        w = d0.copy()
        w[z.replace("z", "x")] = w[z.replace("z", "y")] = 1
        z_out = compute(w, ops)
        [i] = [i for i in range(len(zs)) if 2**i == z_out]
        if int(z[1:]) != i:
            swapped += [z, f"z{i:02d}"]
        if len(swapped) == n_swapped:
            break

# https://en.wikipedia.org/wiki/Adder_(electronics)#Ripple-carry_adder
or_operands = {i0 for i0, op, i1, out in ops if op == "OR"}
or_operands |= {i1 for i0, op, i1, out in ops if op == "OR"}
and_operands = {i0 for i0, op, i1, out in ops if op == "AND"}
and_operands |= {i1 for i0, op, i1, out in ops if op == "AND"}
for i0, op, i1, out in ops:
    if len(swapped) == n_swapped:
        break
    if i0 == "x00" and i1 == "y00":
        continue
    elif out[0] == "z" and out != zs[-1] and op != "XOR":
        swapped.append(out)
    elif out[0] != "z" and i0[0] != "x" and i1[0] != "y" and op == "XOR":
        swapped.append(out)
    elif i0[0] == "x" and i1[0] == "y" and op == "AND" and out not in or_operands:
        swapped.append(out)
    elif i0[0] == "x" and i1[0] == "y" and op == "XOR" and out not in and_operands:
        swapped.append(out)
print("answer_b:", ",".join(sorted(swapped)))



"""
--- Day 23: A Long Walk ---
https://adventofcode.com/2023/day/23
"""
from aocd import data

from aoc_wim.zgrid import ZGrid


grid = ZGrid(data)
nodes = {}
dzs = dict(zip("^>v<", [-1j, 1, 1j, -1]))
for z in grid.z(".", first=False):
    if sum(grid.get(z + dz) in dzs for dz in dzs.values()) >= 2:
        nodes[z] = {}


def find_next_node(z0, dz0):
    path = [z0, z0 + dz0]
    while path[-1] not in nodes:
        [z1] = [z for z in grid.near(path[-1]) if z != path[-2] and grid.get(z, "#") != "#"]
        path.append(z1)
    return path[-1], path[-2] - path[-1], len(path) - 1


start, dzS, dS = find_next_node(1, 1j)
grid[start + dzS] = "#"
end, dzE, dE = find_next_node(grid.bottom_right - 1, -1j)
grid[end + dzE] = "#"

for z0, branches in nodes.items():
    for dz0 in dzs.values():
        g = grid.get(z0 + dz0)
        if g in dzs and dz0 not in branches:
            z1, dz1, dist = find_next_node(z0, dz0)
            branches[dz0] = z1, dist
            nodes[z1][dz1] = z0, dist
perimeter = {z for z, branches in nodes.items() if len(branches) < 4}
node_mask = {z: 1 << i for i, z in enumerate(nodes)}


def longest_path(part="a"):
    dists = []
    stack = [(start, node_mask[start], 0)]
    while stack:
        z0, visited, d0 = stack.pop()
        if z0 == end:
            dists.append(d0)
        branches = nodes[z0]
        for g, dz in dzs.items():
            if dz in branches:
                z1, dist = branches[dz]
                if node_mask[z1] & visited:
                    continue
                if grid[z0 + dz] != g:
                    if part == "a" or {z0, z1} <= perimeter:
                        continue
                stack.append((z1, visited | node_mask[z1], d0 + dist))
    return dS + max(dists) + dE


print("answer_a:", longest_path(part="a"))
print("answer_b:", longest_path(part="b"))



"""
--- Day 19: Go With The Flow ---
https://adventofcode.com/2018/day/19
"""
import math

from aocd import data

from aoc_wim.aoc2018.q16 import all_ops


def parsed(data):
    lines = data.splitlines()
    ip = int(lines.pop(0).split()[1])
    lines = [s.split() for s in lines]
    lines = [[a, int(b), int(c), int(d)] for a, b, c, d in lines]
    return ip, lines


funcs = all_ops()


class Comp:
    def __init__(self, ip, d, r0, hacked):
        self.hacked = hacked
        self.ip = ip
        self.i = 0
        self.d = d
        self.r = [0] * 6
        self.r[0] = r0

    def step(self):
        opname, a, b, c = self.d[self.i]
        op = funcs[opname]
        self.r[self.ip] = self.i
        op(self.r, a, b, c)
        self.r[self.ip] += 1
        self.i = self.r[self.ip]
        if self.hacked and self.i == 1 and max(self.r) > 5:
            self.r[0] = sum(divisors(max(self.r)))
            [][0]


def compute(data, r0=0, hack=False):
    ip, lines = parsed(data)
    comp = Comp(ip, lines, r0=r0, hacked=hack)
    while True:
        try:
            comp.step()
        except IndexError:
            break
    return comp.r[0]


def divisors(n):
    divs = {1, n}
    for i in range(2, int(math.sqrt(n)) + 1):
        if not n % i:
            divs |= {i, n // i}
    return divs


print("answer_a:", compute(data, r0=0, hack=True))
print("answer_b:", compute(data, r0=1, hack=True))



"""
--- Day 24: Blizzard Basin ---
https://adventofcode.com/2022/day/24
"""
from aocd import data

from aoc_wim.zgrid import ZGrid


grid = ZGrid(data)
grid.offset(-1-1j)
h = grid.height - 2
w = grid.width - 2
min_depth = h + w

bu = grid.z("^", first=False)
br = grid.z(">", first=False)
bd = grid.z("v", first=False)
bl = grid.z("<", first=False)


empty = ZGrid({z: ".#"[g == "#"] for z, g in grid.items()})
z_start = -1j
z_end = w - 1 + h * 1j


rep = dict.fromkeys("^>v<", "2")
rep["2"] = "3"
rep["3"] = "4"


def bliz(t):
    g = ZGrid(empty)
    u_y = [(int(z.imag) - t) % h for z in bu]
    r_x = [(int(z.real) + t) % w for z in br]
    d_y = [(int(z.imag) + t) % h for z in bd]
    l_x = [(int(z.real) - t) % w for z in bl]
    u = [z.real + 1j * y for z, y in zip(bu, u_y)]
    r = [x + 1j * z.imag for z, x in zip(br, r_x)]
    d = [z.real + 1j * y for z, y in zip(bd, d_y)]
    l = [x + 1j * z.imag for z, x in zip(bl, l_x)]
    g.update(dict.fromkeys(u, "^"))
    g.update({z: rep.get(g[z], ">") for z in r})
    g.update({z: rep.get(g[z], "v") for z in d})
    g.update({z: rep.get(g[z], "<") for z in l})
    return g


def fill(z, targets):
    zs = {z}
    m = -1
    while targets:
        target = targets.pop()
        while target not in zs:
            m += 1
            g = bliz(m)
            zs = {z for z0 in zs for z in g.near(z0, n=5) if g.get(z) == "."}
        zs = {target}
    return m


print("answer_a:", fill(z=z_start, targets=[z_end]))
print("answer_b:", fill(z=z_start, targets=[z_end, z_start, z_end]))



"""
--- Day 15: Warehouse Woes ---
https://adventofcode.com/2024/day/15
"""
from aocd import data

from aoc_wim.zgrid import ZGrid


warehouse, path = data.split("\n\n")
grid = ZGrid(warehouse)
dzs = [grid.dzs[x] for x in path.replace("\n", "")]

z = grid.z("@")
grid[z] = "."
for dz in dzs:
    val = grid[z + dz]
    if val == "#":
        continue
    elif val == ".":
        z += dz
    elif val == "O":
        zs = [z + dz]
        while grid.get(zs[-1] + dz) == "O":
            zs.append(zs[-1] + dz)
        if grid.get(zs[-1] + dz) == "#":
            continue
        zs.append(zs[-1] + dz)
        grid[zs[0]] = "."
        grid.update(dict.fromkeys(zs[1:], "O"))
        z += dz

a = int(sum(z.real + 100*z.imag for z in grid.z("O", first=False)))
print("answer_a:", a)


class Box:
    def __init__(self, z):
        self.z = z

    def __contains__(self, z):
        return z == self.z or z - 1 == self.z

    def can_move(self, dz):
        if dz == -1 and grid[self.z + dz] == "#":
            return False
        if dz == 1 and grid[self.z + 1 + dz] == "#":
            return False
        if dz in (-1j, 1j) and (grid[self.z + dz] == "#" or grid[self.z + 1 + dz] == "#"):
            return False
        if dz == -1:
            for box in boxes:
                if self.z + dz in box:
                    return box.can_move(dz)
            return True
        if dz == 1:
            for box in boxes:
                if self.z + 1 + dz in box:
                    return box.can_move(dz)
            return True
        box2 = [x for x in boxes if self.z + dz in x or self.z + 1 + dz in x]
        return all(b.can_move(dz) for b in box2)

    def move(self, dz):
        if dz == -1:
            for box in boxes:
                if self.z + dz in box:
                    box.move(dz)
        elif dz == 1:
            for box in boxes:
                if self.z + 1 + dz in box:
                    box.move(dz)
        else:
            for box in boxes:
                if self.z + dz in box or self.z + 1 + dz in box:
                    box.move(dz)
        self.z += dz


remap = {
    "#": "##",
    "O": "[]",
    ".": "..",
    "@": "@.",
}
grid = ZGrid(warehouse.translate(str.maketrans(remap)))
boxes = []
for z_, g in grid.items():
    if g == "[":
        boxes.append(Box(z_))
        grid[z_] = "."
    elif g == "]":
        grid[z_] = "."
    elif g == "@":
        z = z_
        grid[z_] = "."

for dz in dzs:
    if grid[z + dz] == "#":
        continue
    box = [x for x in boxes if z + dz in x]
    if box:
        [box] = box
        if not box.can_move(dz):
            continue
        box.move(dz)
    z += dz

b = int(sum(box.z.real + 100*box.z.imag for box in boxes))
print("answer_b:", b)



"""
--- Day 18: Boiling Boulders ---
https://adventofcode.com/2022/day/18
"""
from itertools import product

import networkx as nx
from aocd import data


def adj6(x, y, z):
    return [
        (x - 1, y, z),
        (x + 1, y, z),
        (x, y - 1, z),
        (x, y + 1, z),
        (x, y, z - 1),
        (x, y, z + 1),
    ]


cubes = {tuple(map(int, line.split(","))) for line in data.splitlines()}

graph = nx.Graph()
m = max(s for cube in cubes for s in cube) + 2
for c in product(range(-1, m), repeat=3):
    if c not in cubes:
        graph.add_node(c)

for c0 in graph:
    for c1 in adj6(*c0):
        if c1 in graph:
            graph.add_edge(c0, c1)

water = max(nx.connected_components(graph), key=len)

print("answer_a:", sum([c1 not in cubes for c0 in cubes for c1 in adj6(*c0)]))
print("answer_b:", sum([c in cubes for w in water for c in adj6(*w)]))



"""
--- Day 12: Hill Climbing Algorithm ---
https://adventofcode.com/2022/day/12
"""
from aocd import data

from aoc_wim.search import AStar
from aoc_wim.zgrid import ZGrid

grid = ZGrid(data)
start = grid.z("S")
end = grid.z("E")
grid[start] = "a"
grid[end] = "z"


class A𖤐(AStar):
    def adjacent(self, z0):
        return [z for z in grid.near(z0) if ord(grid[z0]) - ord(grid.get(z, "_")) <= 1]

    def target_reached(self, z, target):
        return z == target or grid[z] == target


print("answer_a:", len(A𖤐(end, start).run()) - 1)
print("answer_b:", len(A𖤐(end, "a").run()) - 1)



"""
--- Day 18: Many-Worlds Interpretation ---
https://adventofcode.com/2019/day/18
"""
import string
from itertools import combinations

import networkx as nx
from aocd import data

from aoc_wim.search import AStar
from aoc_wim.zgrid import ZGrid


class Q18AStar(AStar):

    def __init__(self, data, part="a"):
        grid = ZGrid(data)
        pos = ("@",)
        if part == "b":
            if grid.count("@") == 1:
                z0 = grid.z("@")
                if grid.count_near(z0, val=".", n=8, default=".") == 8:
                    pos = ("@0", "@1", "@2", "@3")
                    for z in grid.near(z0, n=5):
                        grid[z] = "#"
                    for p, dz in zip(pos, [-1 - 1j, 1 - 1j, 1 + 1j, -1 + 1j]):
                        grid[z0 + dz] = p
            else:
                pos = ()
                for i, z in enumerate(grid.z("@", first=False)):
                    glyph = f"@{i}"
                    grid[z] = glyph
                    pos += (glyph,)

        graph = grid.graph(extra=frozenset(string.ascii_letters).union(pos))
        self.all_keys = {k for k in graph.extra if k in string.ascii_lowercase}

        # map of the other keys that must be acquired before each key is acquired
        self.obstructions = {k: set() for k in self.all_keys}
        # precached distances between points of interest
        self.kdist = {}
        for k0 in self.all_keys:
            for p in pos:
                try:
                    path = nx.shortest_path(graph, graph.extra[p], graph.extra[k0])
                except nx.NetworkXNoPath:
                    continue
                self.kdist[k0, p] = self.kdist[p, k0] = len(path) - 1
                for p in path[1:-1]:
                    if p in graph.extra.values():
                        k1 = grid[p].lower()
                        self.obstructions[k0].add(k1)

        for k1, k2 in combinations(self.all_keys, 2):
            z1 = graph.extra[k1]
            z2 = graph.extra[k2]
            try:
                d = nx.shortest_path_length(graph, z1, z2)
            except nx.NetworkXNoPath:
                continue
            self.kdist[k1, k2] = self.kdist[k2, k1] = d

        state0 = pos, frozenset()
        AStar.__init__(self, state0, target=None)

    def cost(self, state0, state1):
        if state1 is None:
            return 0
        pos0, keys0 = state0
        pos1, keys1 = state1
        [(k0, k1)] = [(k0, k1) for k0, k1 in zip(pos0, pos1) if k0 != k1]
        return self.kdist[k0, k1]

    heuristic = cost

    def target_reached(self, current_state, target):
        _, keys = current_state
        return keys == self.all_keys

    def adjacent(self, state):
        pos, keys = state
        for i, p in enumerate(pos):
            for key, obstruction in self.obstructions.items():
                if key not in keys and keys >= obstruction and (p, key) in self.kdist:
                    yield pos[:i] + (key,) + pos[i+1:], keys.union(key)


astar = Q18AStar(data, part="a")
path_a = astar.run()
print("answer_a:", astar.gscore[astar.target])

astar = Q18AStar(data, part="b")
path_b = astar.run()
print("answer_b:", astar.gscore[astar.target])



"""
--- Day 3: Mull It Over ---
https://adventofcode.com/2024/day/3
"""
import parse
from aocd import data


pat = "mul({:d},{:d})"
a = sum(x * y for x, y in parse.findall(pat, data))
print("answer_a:", a)

while "don't()" in data:
    data, tail = data.split("don't()", 1)
    if "do()" in tail:
        data += tail.split("do()", 1)[1]
b = sum(x * y for x, y in parse.findall(pat, data))
print("answer_b:", b)



import argparse
import ast
import shutil
import sys
import tempfile
import webbrowser
from datetime import datetime
from pathlib import Path
from textwrap import dedent

import aocd
from aocd.models import Puzzle
from aocd.utils import AOC_TZ
from aocd.utils import blocker


here = Path(__file__).parent.resolve()


def get_module_docstring(path):
    """get a .py file docstring, without actually executing the file"""
    with open(path) as f:
        return ast.get_docstring(ast.parse(f.read()))


def set_module_docstring(path, text):
    """update docstring of file at path with text, cautiously"""
    prev = get_module_docstring(path)
    if prev == text:
        # nothing to do
        return
    if prev is not None:
        print(repr(text))
        print(repr(prev))
        raise Exception(f"refusing to clobber existing docstring on {path}")
    content = Path(path).read_text()
    Path(path).write_text(f'"""\n{text}\n"""\n{content}')


def set_docstrings(files=()):
    """puts name and link into puzzle source files"""
    for file in files or here.glob("aoc*/q*.py"):
        day = int(file.name[1:3])
        year = int(file.parent.name[3:])
        puzzle = Puzzle(year, day)
        docstring = f"--- Day {day}: {puzzle.title} ---\n"
        docstring += puzzle.url
        set_module_docstring(file, docstring)


def split_trailing_comments(lines):
    extra = []
    while lines and (not lines[-1].strip() or lines[-1].startswith("#")):
        extra.append(lines.pop())
    if len(lines) and "#" in lines[-1]:
        line, comment = lines[-1].split("#", 1)
        lines[-1] = line.strip()
        extra.append(comment.strip())
    if len(lines) > 1 and "#" in lines[-2]:
        line, comment = lines[-2].split("#", 1)
        lines[-2] = line.strip()
        extra.append(comment.strip())
    extra = [x.strip() for x in extra if x.strip()]
    return extra


def parse_extra_context(extra):
    result = {}
    for line in extra:
        equals = line.count("=")
        commas = line.count(",")
        if equals and equals == commas + 1:
            for part in line.split(","):
                k, v = part.strip().split("=")
                k = k.strip()
                v = v.strip()
                try:
                    v = ast.literal_eval(v)
                except ValueError:
                    pass
                if k in result:
                    raise NotImplementedError(f"Duplicate key {k!r}")
                result[k] = v
    return result


def start():
    """init a new source file at ./aocYYYY/qDD.py"""
    aoc_now = datetime.now(tz=AOC_TZ)
    years = range(2015, aoc_now.year + int(aoc_now.month >= 11))
    days = range(1, 26)
    parser = argparse.ArgumentParser(description="init current day")
    parser.add_argument(
        "day",
        nargs="?",
        type=int,
        default=min(aoc_now.day, 25) if aoc_now.month == 12 else 1,
        help="1-25 (default: %(default)s)",
    )
    parser.add_argument(
        "year",
        nargs="?",
        type=int,
        default=years[-1],
        help="2015-%(default)s (default: %(default)s)",
    )
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="create the template file even if source already exists",
    )
    args = parser.parse_args()
    if args.day in years and args.year in days:
        # be forgiving
        args.day, args.year = args.year, args.day
    if args.day not in days or args.year not in years:
        parser.print_usage()
        parser.exit(1)
    year = args.year
    day = args.day
    here = Path(__file__).parent
    src = here / f"aoc{year}" / f"q{day:02d}.py"
    block = False
    if day < 25 and year == aoc_now.year and day == aoc_now.day:
        if src.exists() and str(day) not in sys.argv:
            day += 1
            block = True
            src = here / f"aoc{year}" / f"q{day:02d}.py"
    if src.exists():
        if not args.force:
            sys.exit(f"{src} already exists! (use -f to overwrite)")
        shutil.copy2(src, tempfile.gettempdir())
    if day == 1 and not src.parent.is_dir():
        src.parent.mkdir()
        src.parent.joinpath("__init__.py").touch()
    src.write_text(
        dedent(
            """\
            from aocd import data
            from aoc_wim.autoparse import parsed
            from collections import Counter, defaultdict, deque
            d = parsed(data)
            if d != data:
                print(d)
                print(f"{parsed.n_bytes} bytes/{parsed.n_lines} lines, parsed by {parsed.parser}")

            # import numpy as np
            # import networkx as nx
            # from aoc_wim.zgrid import ZGrid; grid = ZGrid(data)
            # from aoc_wim import stuff

            # import logging; logging.basicConfig(level=logging.DEBUG)





            print("answer_a:", )
            print("answer_b:", )

            # from aocd import submit; submit(a)
            """
        )
    )
    test = here.parent / "tests" / str(year) / str(day).zfill(2) / "0.txt"
    if not test.exists():
        test.parent.mkdir(parents=True, exist_ok=True)
        test.write_text("\n-\n-\n")
    if block:
        blocker()
    data = aocd.get_data(day=day, year=year, block=True)
    print(data)
    set_docstrings([src])
    puzzle = Puzzle(year, day)
    [example] = puzzle.examples
    if test.read_text() == "\n-\n-\n":
        test.write_text(
            f"{example.input_data}\n"
            f"{example.answer_a or '-'}\n"
            f"{example.answer_b or '-'}\n"
        )
    webbrowser.open(puzzle.url)



"""
--- Day 16: Dragon Checksum ---
https://adventofcode.com/2016/day/16
"""
from aocd import data
from aocd import extra


def dragon(a):
    return a + "0" + "".join([{"0": "1", "1": "0"}[x] for x in a[::-1]])


def pad(s, n):
    return pad(dragon(s), n) if len(s) < n else s[:n]


def checksum(s):
    return "".join(["01"[a == b] for a, b in zip(s[0::2], s[1::2])])


def r_checksum(s):
    s = checksum(s)
    return s if len(s) % 2 else r_checksum(s)


def f(data, length):
    return r_checksum(pad(data, length))


if __name__ == "__main__":
    disk_length = extra.get("disk_length", 272)
    print("answer_a:", f(data, disk_length))
    if not extra:
        print("answer_b:", f(data, 35651584))



"""
--- Day 17: Clumsy Crucible ---
https://adventofcode.com/2023/day/17
"""
from aocd import data

from aoc_wim.search import AStar
from aoc_wim.zgrid import ZGrid


class Q17AStar(AStar):
    def __init__(self, state0, min_tail, max_tail):
        self.min_tail = min_tail
        self.max_tail = max_tail
        super().__init__(state0, None)

    def cost(self, current_state, next_state):
        return grid[next_state[-1]]

    def adjacent(self, state0):
        tail, dz, z0 = state0
        if tail < self.max_tail and z0 + dz in grid:
            yield tail + 1, dz, z0 + dz
        if self.min_tail <= tail:
            for turn in grid.turn_left, grid.turn_right:
                z = z0 + dz * turn
                if z in grid:
                    yield 1, dz * turn, z
        if z0 == 0 and self.min_tail:
            yield 1, 1j, z0 + 1j

    def target_reached(self, current_state, target):
        tail, dz, z = current_state
        return z == bottom_right and tail >= self.min_tail


grid = ZGrid(data, transform=int)
s0 = (0, 1, 0)  # state vector: tail-length, direction, position
bottom_right = grid.bottom_right  # target

astar = Q17AStar(s0, min_tail=0, max_tail=3)
astar.run()
print("answer_a:", astar.gscore[astar.target])

bstar = Q17AStar(s0, min_tail=4, max_tail=10)
bstar.run()
print("answer_b:", bstar.gscore[bstar.target])



"""
--- Day 4: Passport Processing ---
https://adventofcode.com/2020/day/4
"""
from aocd import data
from marshmallow import INCLUDE
from marshmallow import Schema
from marshmallow.fields import Integer
from marshmallow.fields import String
from marshmallow.validate import OneOf
from marshmallow.validate import Range
from marshmallow.validate import Regexp
from marshmallow.validate import ValidationError
from marshmallow.validate import Validator


class Height(Validator):
    def __call__(self, value):
        if value.endswith("cm"):
            Range(min=150, max=193)(int(value[:-2]))
        elif value.endswith("in"):
            Range(min=59, max=76)(int(value[:-2]))
        else:
            raise ValidationError("height must be cm or in")
        return value


class PassportSchema(Schema):
    byr = Integer(validate=Range(min=1920, max=2002))
    iyr = Integer(validate=Range(min=2010, max=2020))
    eyr = Integer(validate=Range(min=2020, max=2030))
    hgt = String(validate=Height())
    hcl = String(validate=Regexp(r"^#[0-9a-f]{6}$"))
    ecl = String(validate=OneOf("amb blu brn gry grn hzl oth".split()))
    pid = String(validate=Regexp(r"^[0-9]{9}$"))

    class Meta:
        unknown = INCLUDE


schema = PassportSchema()
a = b = 0
for block in data.split("\n\n"):
    passport_data = dict(x.split(":") for x in block.split())
    if not (passport_data.keys() >= schema.fields.keys()):
        continue
    a += 1
    errors = schema.validate(passport_data)
    b += not errors

print("answer_a:", a)
print("answer_b:", b)



"""
--- Day 10: Balance Bots ---
https://adventofcode.com/2016/day/10
"""
from math import prod

from aocd import data
from aocd import extra
from parse import parse


class Bot:
    def __init__(self, name, low_to=None, high_to=None):
        self.name = name
        self.low_to = low_to
        self.high_to = high_to
        self.chips = []
        self.part_b = False

    def receive(self, chip):
        self.chips.append(chip)
        if len(self.chips) == 2:
            self.send()

    def send(self):
        low, high = sorted(self.chips)
        self.chips = []
        self.low_to.receive(low)
        self.high_to.receive(high)
        if (low, high) == (lo, hi):
            self.part_b = True


receives = []
receivers = {}
lines = data.splitlines()
hi = extra.get("chip1", 61)
lo = extra.get("chip2", 17)
for line in lines:
    if line.startswith("value"):
        tup = parse("value {:d} goes to {}", line).fixed
        receives.append(tup)
    else:
        result = parse("{} gives low to {} and high to {}", line)
        bot_name, low_to, high_to = result.fixed
        bot = receivers.get(bot_name, Bot(bot_name))
        bot.low_to = receivers.get(low_to, Bot(low_to))
        bot.high_to = receivers.get(high_to, Bot(high_to))
        for receiver in bot, bot.low_to, bot.high_to:
            receivers[receiver.name] = receiver

for n, bot_name in receives:
    receivers[bot_name].receive(n)
outputs = [receivers[f"output {o}"].chips for o in (0, 1, 2)]
[bot_a] = [v for v in receivers.values() if v.part_b]
print("answer_a:", int(bot_a.name.lstrip("bot ")))
print("answer_b:", prod([n for [n] in outputs]))



"""
--- Day 7: No Space Left On Device ---
https://adventofcode.com/2022/day/7
"""
from collections import Counter
from pathlib import Path

from aocd import data

cwd = root = Path("/")
dirs = Counter()

for line in data.splitlines():
    match line.split():
        case ["$", "cd", "/"]: cwd = root
        case ["$", "cd", ".."]: cwd = cwd.parent
        case ["$", "cd", name]: cwd /= name
        case [s, name] if s.isdigit():
            for p in [cwd, *cwd.parents]:
                dirs[p] += int(s)

rmbytes = dirs[root] - 70_000_000 + 30_000_000
a = sum(v for v in dirs.values() if v <= 100_000)
b = min(v for v in dirs.values() if v >= rmbytes)

print("answer_a:", a)
print("answer_b:", b)



"""
--- Day 8: Playground ---
https://adventofcode.com/2025/day/8
"""

from aocd import data
from aocd import extra

ns = [int(x) for x in data.replace(",", " ").split()]
xs = ns[0::3]
ys = ns[1::3]
zs = ns[2::3]
[n] = {len(xs), len(ys), len(zs)}
d2 = {}
for i in range(n):
    for j in range(i + 1, n):
        dx = xs[i] - xs[j]
        dy = ys[i] - ys[j]
        dz = zs[i] - zs[j]
        d2[frozenset([i, j])] = dx**2 + dy**2 + dz**2
d2 = {k: d2[k] for k in sorted(d2, key=d2.get, reverse=True)}


def merge(d2, components):
    (i, j), _d = d2.popitem()
    component = components[i] | components[j]
    for x in component:
        components[x] = component
    return i, j


n_pairs = extra.get("n_pairs", 1000)
components = {i: {i} for i in range(n)}
for _ in range(n_pairs):
    merge(d2, components)

unique_components = {frozenset(x) for x in components.values()}
sizes = sorted([len(x) for x in unique_components], reverse=True)
print("answer_a:", sizes[0] * sizes[1] * sizes[2])

while len({id(x) for x in components.values()}) > 1:
    i, j = merge(d2, components)
print("answer_b:", xs[i] * xs[j])



"""
--- Day 1: Historian Hysteria ---
https://adventofcode.com/2024/day/1
"""
from collections import Counter

from aocd import data

ns = [int(n) for n in data.split()]
xs = sorted(ns[0::2])
ys = sorted(ns[1::2])
a = sum(abs(x - y) for x, y in zip(xs, ys))
rcount = Counter(ys)
b = sum(x * rcount[x] for x in xs)
print("answer_a:", a)
print("answer_b:", b)



"""
--- Day 12: JSAbacusFramework.io ---
https://adventofcode.com/2015/day/12
"""
import json
import re

from aocd import data


def numbers(s):
    return re.findall(r"-?\d+", s)


def rsum(data):
    match data:
        case int(): return data
        case list(): return sum(rsum(x) for x in data)
        case dict():
            vals = list(data.values())
            return 0 if "red" in vals else rsum(vals)
        case _: return 0


print("answer_a:", sum(map(int, numbers(data))))
print("answer_b:", rsum(json.loads(data)))



import json
import pathlib

import pytest

from aoc_wim import plugin
from aoc_wim.util import parse_extra_context
from aoc_wim.util import split_trailing_comments


here = pathlib.Path(__file__).parent
input_files = sorted(here.glob("20*/*/*.txt"))


def path2id(input_file):
    rel_path = input_file.relative_to(here)
    result = str(rel_path)
    titles = getattr(pytest, "puzzle_titles", {})
    title = titles.get(str(rel_path.parent))
    if title:
        result += f" : {title}"
    return result


@pytest.mark.parametrize("input_file", input_files, ids=path2id)
def test_example(input_file, monkeypatch, request):
    # example input files are in ./YYYY/dd/fname.txt
    *pre, year, day, fname = input_file.parts
    year = int(year)
    day = int(day)
    # the head of each example file is an input data
    # the last two lines are part a and part b correct answers
    # there may be trailing comments after the answers
    lines = input_file.read_text(encoding="utf-8").splitlines()
    extra = split_trailing_comments(lines)
    extra = parse_extra_context(extra)
    if extra:
        monkeypatch.setenv("AOCD_EXTRA", json.dumps(extra))
    if len(lines) < 3:
        pytest.fail(f"test data {input_file} is malformed")
    *lines, expected_answer_a, expected_answer_b = lines
    input_data = "\n".join(lines).rstrip()
    # patch out aocd
    monkeypatch.setattr("aocd.data", input_data)
    # invoke the entrypoint with a controlled input
    actual_answer_a, actual_answer_b = plugin(year, day, input_data)
    # verify correct answers returned. sometimes a dash (-) may be used to indicate
    # an example input where the answer was provided for only one of the two parts
    if expected_answer_a != "-":
        assert actual_answer_a == expected_answer_a, f"{actual_answer_a=} {expected_answer_a=}"
    if expected_answer_b != "-" and not request.config.getoption("--part-a-only"):
        assert actual_answer_b == expected_answer_b, f"{actual_answer_b=} {expected_answer_b=}"



"""
--- Day 12: Hot Springs ---
https://adventofcode.com/2023/day/12
"""
from functools import cache

from aocd import data


@cache
def fits(ss, cs, h=0):
    if not any(cs):
        return "#" not in ss
    if not ss:
        return 0
    c, cs = cs[-1], cs[:-1]
    s, ss = ss[-1], ss[:-1]
    if not c:
        return (s != "#") and fits(ss, cs, 0)
    if h:
        return (s != ".") and fits(ss, (*cs, c-1), 1)
    if s == "#":
        return fits(ss, (*cs, c - 1), 1)
    if s == ".":
        return fits(ss, (*cs, c), 0)
    if s == "?":
        return fits(ss, (*cs, c), 0) + fits(ss, (*cs, c - 1), 1)


a = b = 0
for line in data.splitlines():
    template_a, counts = line.split()
    counts_a = *map(int, counts.split(",")),
    template_b = "?".join([template_a]*5)
    counts_b = counts_a*5
    a += fits(template_a, counts_a)
    b += fits(template_b, counts_b)
    fits.cache_clear()

print("answer_a:", a)
print("answer_b:", b)



"""
--- Day 5: Supply Stacks ---
https://adventofcode.com/2022/day/5
"""
from aocd import data
from parse import parse

crates, instructions = data.split("\n\n")
cols0 = {}
for *row, last in zip(*crates.split("\n")):
    if last.isdigit():
        cols0[int(last)] = "".join(x for x in row if x.isupper())

moves = []
for line in instructions.splitlines():
    n, p0, p1 = parse("move {:d} from {:d} to {:d}", line).fixed
    moves.append((n, p0, p1))

for part in "ab":
    step = -1 if part == "a" else None
    cols = cols0.copy()
    for n, p0, p1 in moves:
        cols[p1] = cols[p0][:n][::step] + cols[p1]
        cols[p0] = cols[p0][n:]
    result = "".join([v0 for v0, *_ in cols.values()])
    print(f"answer_{part}:", result)



"""
--- Day 21: Dirac Dice ---
https://adventofcode.com/2021/day/21
"""
from collections import Counter

from aocd import data


def play_deterministic(p1, p2):
    pos = [p1, p2]
    score = [0, 0]
    rolls = 0
    while True:
        for i in 0, 1:
            rolls += 3
            pos[i] += 3 * (rolls - 1)
            pos[i] = (pos[i] - 1) % 10 + 1
            score[i] += pos[i]
            if score[i] >= 1000:
                return rolls * score[i - 1]


def wins(p1, p2, s1, s2, memo, rolls):
    if s2 >= 21:
        return 0, 1
    if (p1, p2, s1, s2) not in memo:
        t1 = t2 = 0
        for roll, count in rolls.items():
            next_p1 = (p1 + roll - 1) % 10 + 1
            wins2, wins1 = wins(p2, next_p1, s2, s1 + next_p1, memo, rolls)
            t1 += wins1 * count
            t2 += wins2 * count
        memo[p1, p2, s1, s2] = t1, t2
    return memo[p1, p2, s1, s2]


def play_quantum(p1, p2):
    r = [1, 2, 3]
    rolls = Counter([r1 + r2 + r3 for r1 in r for r2 in r for r3 in r])
    wins1, wins2 = wins(p1, p2, 0, 0, {}, rolls)
    return max(wins1, wins2)


p1, p2 = [int(line.split()[-1]) for line in data.splitlines()]
print("answer_a:", play_deterministic(p1, p2))
print("answer_b:", play_quantum(p1, p2))



"""
--- Day 22: Monkey Market ---
https://adventofcode.com/2024/day/22
"""
from collections import Counter

import numpy as np
from aocd import data


n = np.array([*map(int, data.split())])
N = np.empty((2001, len(n)), dtype=int)
N[0] = n
for i in range(1, 2001):
    n ^= (n << 6) & 0xffffff
    n ^= (n >> 5)
    n ^= (n << 11) & 0xffffff
    N[i] = n
print("answer_a:", n.sum())

N %= 10
dN = N[1:] - N[:-1]
dN3 = dN[0:-3] << 15
dN2 = dN[1:-2] << 10
dN1 = dN[2:-1] << 5
key = dN[3:] + dN3 + dN2 + dN1

bananas = Counter()
for i in range(len(n)):
    seq = key.T[i][::-1].tolist()
    price = N[4:].T[i][::-1].tolist()
    bananas += dict(zip(seq, price))
print("answer_b:", max(bananas.values()))



"""
--- Day 1: Trebuchet?! ---
https://adventofcode.com/2023/day/1
"""
import regex as re
from aocd import data

ns = "one, two, three, four, five, six, seven, eight, nine".split(", ")
d = dict(zip(ns, "123456789"))
a = b = 0
pat_a = re.compile(r"\d")
pat_b = re.compile(rf"(\d|{'|'.join(ns)})")
for line in data.splitlines():
    nums_a = [int(x) for x in pat_a.findall(line)] or [0]
    nums_b = [int(d.get(x, x)) for x in pat_b.findall(line, overlapped=True)] or [0]
    a += nums_a[0] * 10 + nums_a[-1]
    b += nums_b[0] * 10 + nums_b[-1]

print("answer_a:", a)
print("answer_b:", b)



"""
--- Day 6: Memory Reallocation ---
https://adventofcode.com/2017/day/6
"""
from itertools import count

import numpy as np
from aocd import data


def part_ab(data):
    a = np.fromstring(data, sep=" ", dtype=int)
    n = len(a)
    seen = {}
    for i in count():
        t = tuple(a)
        if t in seen:
            return i, i - seen[t]
        seen[t] = i
        max_pos = a.argmax()
        q, r = divmod(a[max_pos], n)
        a[max_pos] = 0
        a += np.roll([q + 1] * r + [q] * (n - r), max_pos + 1)


a, b = part_ab(data)
print("answer_a:", a)
print("answer_b:", b)



"""
--- Day 24: Lobby Layout ---
https://adventofcode.com/2020/day/24
"""
import re

from aocd import data

from aoc_wim import zgrid


pat = r"(se|sw|ne|nw|e|w)"
grid = zgrid.ZGrid()
for i, line in enumerate(data.splitlines()):
    steps = re.findall(pat, line)
    z = sum([zgrid.hexH[s] for s in steps])
    grid[z] = not grid.get(z)
    # grid.draw_hex(glyph=0, orientation="H", clear=True, title=f" flip {i} ")
print("answer_a:", grid.count(1))


for day in range(1, 101):
    z_black = {z for z, v in grid.items() if v}
    tiles_to_visit = z_black.union(*[grid.near(z, n=6) for z in z_black])
    tiles_to_update = {}
    for z in tiles_to_visit:
        n_black = grid.count_near(z, n=6, val=1)
        if grid.get(z) and n_black == 0 or n_black > 2:
            tiles_to_update[z] = 0
        elif n_black == 2:
            tiles_to_update[z] = 1
    grid.d.update(tiles_to_update)
    if day <= 10 or day % 10 == 0:
        print(f"Day {day}:", grid.count(1))
    if day == 10:
        print()
print("answer_b:", grid.count(1))



"""
--- Day 24: Never Tell Me The Odds ---
https://adventofcode.com/2023/day/24
"""
import itertools as it
from dataclasses import dataclass

import z3
from aocd import data
from aocd import extra


@dataclass
class Hailstone:
    x: int
    y: int
    z: int
    dx: int
    dy: int
    dz: int

    @classmethod
    def fromline(cls, line):
        return cls(*map(int, line.replace("@", ",").split(",")))


def intersect2d(h0, h1):
    denom = h0.dx * h1.dy - h0.dy * h1.dx
    if denom == 0:
        return float("inf"), float("inf")
    f = h0.y * h0.dx - h0.x * h0.dy
    g = h1.x * h1.dy - h1.y * h1.dx
    px = (f * h1.dx + g * h0.dx) / denom
    py = (f * h1.dy + g * h0.dy) / denom
    return px, py


test_area_min = extra.get("test_area_min", 200_000_000_000_000)
test_area_max = extra.get("test_area_max", 400_000_000_000_000)
a = 0
hailstones = [Hailstone.fromline(line) for line in data.splitlines()]
for h0, h1 in it.combinations(hailstones, 2):
    px, py = intersect2d(h0, h1)
    if test_area_min <= px <= test_area_max and test_area_min <= py <= test_area_max:
        if (px > h0.x and h0.dx > 0) or (px < h0.x and h0.dx < 0):
            if (px > h1.x and h1.dx > 0) or (px < h1.x and h1.dx < 0):
                a += 1
print("answer_a:", a)


# https://www.reddit.com/r/adventofcode/comments/18pnycy/comment/kepu26z/?utm_source=share&utm_medium=web3x&utm_name=web3xcss&utm_term=1&utm_content=share_button ?
x, y, z, dx, dy, dz = z3.Ints("x y z dx dy dz")
ts = z3.Ints("t0 t1 t2")
s = z3.Solver()
for t, h in zip(ts, hailstones):
    s.add(t >= 0)
    s.add(h.x + t * h.dx == x + t * dx)
    s.add(h.y + t * h.dy == y + t * dy)
    s.add(h.z + t * h.dz == z + t * dz)
s.check()
m = s.model()
b = m[x].as_long() + m[y].as_long() + m[z].as_long()
print("answer_b:", b)



"""
--- Day 5: Print Queue ---
https://adventofcode.com/2024/day/5
"""
from collections import defaultdict
from functools import cmp_to_key

from aocd import data


rules, updates = data.split("\n\n")
rules = [x.split("|") for x in rules.splitlines()]
updates = [x.split(",") for x in updates.splitlines()]

d = defaultdict(set)
for left, right in rules:
    d[left].add(right)


def cmp(p1, p2):
    if p2 in d.get(p1, []):
        return -1
    if p1 in d.get(p2, []):
        return 1
    return 0


a = b = 0
for update in updates:
    pages = sorted(update, key=cmp_to_key(cmp))
    mid = int(pages[len(pages) // 2])
    if pages == update:
        a += mid
    else:
        b += mid

print("answer_a:", a)
print("answer_b:", b)



"""
--- Day 8: Resonant Collinearity ---
https://adventofcode.com/2024/day/8
"""
import itertools as it
from collections import defaultdict

from aocd import data

from aoc_wim.zgrid import ZGrid

grid = ZGrid(data)
d = defaultdict(list)
for z, g in grid.items():
    if g not in ".#":
        d[g].append(z)

a = set()
b = set()
for zs in d.values():
    for z1, z2 in it.combinations(zs, 2):
        b.update((z1, z2))
        for z, dz in ((z1, z1 - z2), (z2, z2 - z1)):
            ns = []
            while z + dz in grid:
                z += dz
                ns.append(z)
            a.update(ns[:1])
            b.update(ns)

print("answer_a:", len(a))
print("answer_b:", len(b))



"""
--- Day 13: Point of Incidence ---
https://adventofcode.com/2023/day/13
"""
from aocd import data

data = data.translate(str.maketrans("#.", "10"))
a = b = 0
for chunk in data.split("\n\n"):
    nr = [int(line, 2) for line in chunk.splitlines()]
    nc = [int("".join(line), 2) for line in zip(*chunk.splitlines())]
    for ns, f in zip([nr, nc], [100, 1]):
        for i in range(len(ns)):
            p = sum([(x ^ y).bit_count() for x, y in zip(ns[:i][::-1], ns[i:])])
            a += f * i * (p == 0)
            b += f * i * (p == 1)

print("answer_a:", a)
print("answer_b:", b)



"""
--- Day 14: Restroom Redoubt ---
https://adventofcode.com/2024/day/14
"""
import math
import statistics
from collections import Counter
from dataclasses import dataclass

import parse
from aocd import data
from aocd import extra


w = extra.get("width", 101)
h = extra.get("height", 103)


@dataclass
class Bot:
    x: int
    y: int
    dx: int
    dy: int

    @classmethod
    def fromline(cls, line):
        [parsed] = parse.findall("p={:d},{:d} v={:d},{:d}", line)
        return cls(*parsed.fixed)

    def move(self, t=1):
        self.x = (self.x + self.dx * t) % w
        self.y = (self.y + self.dy * t) % h

    @property
    def quadrant(self):
        if self.x == w//2 or self.y == h//2:
            # "Robots that are exactly in the middle (horizontally or vertically) don't
            # count as being in any quadrant"
            return
        return self.x < w//2, self.y < h//2


bots = [Bot.fromline(x) for x in data.splitlines()]
for bot in bots:
    bot.move(100)
quadrants = [bot.quadrant for bot in bots]
quadrant_counts = Counter(quadrants)
del quadrant_counts[None]
a = math.prod(quadrant_counts.values())
print("answer_a:", a)

b = 100
var0 = statistics.pvariance([bot.x + bot.y for bot in bots])
if not extra:
    while True:
        for bot in bots:
            bot.move()
        b += 1
        if statistics.pvariance([bot.x + bot.y for bot in bots]) < var0/2:
            print("answer_b:", b)
            break



"""
--- Day 10: Hoof It ---
https://adventofcode.com/2024/day/10
"""
import networkx as nx
from aocd import data

from aoc_wim.zgrid import ZGrid

grid = ZGrid(data, transform=int)
graph = nx.DiGraph()
for z0, n in grid.items():
    for z in grid.near(z0):
        if grid.get(z, n) - n == 1:
            graph.add_edge(z0, z)

a = b = 0
for head in grid.z(0, first=False):
    for peak in grid.z(9, first=False):
        paths = list(nx.all_simple_paths(graph, head, peak))
        a += bool(paths)
        b += len(paths)

print("answer_a:", a)
print("answer_b:", b)



"""
--- Day 14: Disk Defragmentation ---
https://adventofcode.com/2017/day/14
"""
import numpy as np
from aocd import data
from scipy.ndimage import label

from aoc_wim.aoc2017.q10 import knot_hash

s = "".join(f"{int(knot_hash(f'{data}-{i}'), 16):0128b}" for i in range(128))
print("answer_a:", s.count("1"))
print("answer_b:", label(np.fromiter(s, dtype=int).reshape(128, 128))[1])



"""
--- Day 6: Wait For It ---
https://adventofcode.com/2023/day/6
"""
from math import prod
from math import sqrt

from aocd import data


def quadratic(t, d):
    return t - 1 - 2*int(t/2 - sqrt(t*t - 4*d)/2)


ts, ds = data.splitlines()
ts = ts.split()[1:]
ds = ds.split()[1:]

print("answer_a:", prod(quadratic(int(t), int(d)) for t, d in zip(ts, ds)))
print("answer_b:", quadratic(int("".join(ts)), int("".join(ds))))



"""
--- Day 16: Reindeer Maze ---
https://adventofcode.com/2024/day/16
"""
import itertools as it
from queue import PriorityQueue

from aocd import data

from aoc_wim.zgrid import ZGrid


grid = ZGrid(data)
q = PriorityQueue()
i = it.count()
S, E = grid.z("S"), grid.z("E")
state = (0, next(i), S, 1, [S])
q.put(state)
a = None
score = {}
paths = []
while not q.empty():
    s0, _, z0, dz0, path = q.get()
    if s0 > score.get((z0, dz0), float("inf")):
        continue
    score[z0, dz0] = s0
    if z0 == E:
        if a is None:
            a = s0
        if a == s0:
            paths.append(path)
    for z in grid.near(z0):
        if grid[z] == "#":
            continue
        dz = z - z0
        s = s0 + [1001, 1][dz == dz0]
        q.put((s, next(i), z, dz, path + [z]))

print("answer_a:", a)
print("answer_b:", len(set().union(*paths)))



"""
--- Day 20: Particle Swarm ---
https://adventofcode.com/2017/day/20
"""
from collections import defaultdict
from itertools import batched

from aocd import data
from parse import parse


def key(p):
    return [abs(x) + abs(y) + abs(z) for x, y, z in batched(reversed(p), 3)]


def abs_accel(p):
    px, py, pz, vx, vy, vz, ax, ay, az = p
    a = abs(ax) + abs(ay) + abs(az)
    return a


def pos(p0, t):
    px0, py0, pz0, vx0, vy0, vz0, ax0, ay0, az0 = p0
    px1 = px0 + vx0 * t + (ax0 * t ** 2) // 2
    py1 = py0 + vy0 * t + (ay0 * t ** 2) // 2
    pz1 = pz0 + vz0 * t + (az0 * t ** 2) // 2
    d = abs(px1) + abs(py1) + abs(pz1)
    return d


template = "p=<{:d},{:d},{:d}>, v=<{:d},{:d},{:d}>, a=<{:d},{:d},{:d}>"
particles = [list(parse(template, line).fixed) for line in data.splitlines()]
p_min = min(particles, key=key)
a_min = abs_accel(p_min)
d = {i: pos(p, 1000) for i, p in enumerate(particles) if abs_accel(p) == a_min}
print("answer_a:", min(d, key=d.get))

for _ in range(100):
    d = defaultdict(list)
    for i, p in enumerate(particles):
        d[p[0], p[1], p[2]].append(i)
    kill = set()
    for k, v in d.items():
        if len(v) > 1:
            kill |= set(v)
    for k in sorted(kill, reverse=True):
        del particles[k]
    for p in particles:
        # accelerate
        p[3] += p[6]
        p[4] += p[7]
        p[5] += p[8]
        # translate
        p[0] += p[3]
        p[1] += p[4]
        p[2] += p[5]
print("answer_b:", len(particles))



"""
--- Day 12: Garden Groups ---
https://adventofcode.com/2024/day/12
"""
import networkx as nx
from aocd import data

from aoc_wim.zgrid import ZGrid


grid = ZGrid(data)
graph = nx.Graph()
for z, g in grid.items():
    graph.add_node(z)
    if grid.get(z + 1) == g:
        graph.add_edge(z, z + 1)
    if grid.get(z + 1j) == g:
        graph.add_edge(z, z + 1j)

regions = [frozenset(r) for r in nx.connected_components(graph)]
for r in regions:
    grid.update(dict.fromkeys(r, r))  # label

area = {r: len(r) for r in regions}
perimeter = {r: sum(4 - len(graph[z]) for z in r) for r in regions}
n_sides = dict.fromkeys(regions, 0)
for z, r in grid.items():
    dzs = -1j, 1, 1j, -1, -1j
    for dz0, dz2 in zip(dzs, dzs[1:]):
        rs = [grid.get(z + dz) for dz in (dz0, dz0 + dz2, dz2)]
        n_sides[r] += r is not rs[0] and r is not rs[2]
        n_sides[r] += r is rs[0] is rs[2] and r is not rs[1]

print("answer_a:", sum(area[r] * perimeter[r] for r in regions))
print("answer_b:", sum(area[r] * n_sides[r] for r in regions))



"""
--- Day 1: Secret Entrance ---
https://adventofcode.com/2025/day/1
"""

from aocd import data


a = b = 0
pos = 50
for n in map(int, data.replace("R", "").replace("L", "-").split()):
    d = -1 if n < 0 else 1
    q, n = divmod(n, d * 100)
    b += q
    for i in range(abs(n)):
        pos += d
        b += pos % 100 == 0
    a += pos % 100 == 0
print("answer_a:", a)
print("answer_b:", b)



"""
--- Day 7: Laboratories ---
https://adventofcode.com/2025/day/7
"""

from aocd import data

lines = data.splitlines()
w = len(lines[0])
beams = [0] * w
beams[lines[0].index("S")] += 1
splitters = [{i for i, x in enumerate(line) if x == "^"} for line in lines[2::2]]

a = 0
for xs in splitters:
    new_beams = [0] * w
    for i, count in enumerate(beams):
        if count:
            if i in xs:
                a += 1
                new_beams[i - 1] += count
                new_beams[i + 1] += count
            else:
                new_beams[i] += count
    beams = new_beams
b = sum(beams)
print("answer_a:", a)
print("answer_b:", b)



"""
--- Day 14: Reindeer Olympics ---
https://adventofcode.com/2015/day/14
"""
from aocd import data
from aocd import extra


class Reindeer:
    def __init__(self, line):
        words = line.split()
        self.name = words[0]
        self.run_speed = int(words[3])
        self.stamina = int(words[6])
        self.rest_time = int(words[-2])
        self.distance = 0
        self.points = 0
        self.speed = self.run_speed
        self.dt = self.stamina

    def step(self):
        self.distance += self.speed
        self.dt -= 1
        if self.dt == 0:
            if self.speed == 0:
                self.speed = self.run_speed
                self.dt = self.stamina
            else:
                self.speed = 0
                self.dt = self.rest_time


deers = [Reindeer(line) for line in data.splitlines()]
t = extra.get("t", 2503)
for _ in range(t):
    for deer in deers:
        deer.step()
    max_distance = max([deer.distance for deer in deers])
    for deer in deers:
        if deer.distance == max_distance:
            deer.points += 1

print("answer_a:", max([d.distance for d in deers]))
print("answer_b:", max([d.points for d in deers]))



"""
--- Day 2: Gift Shop ---
https://adventofcode.com/2025/day/2
"""

from aocd import data

ranges = [xy.split("-") for xy in data.split(",")]
ranges = [range(int(x), int(y) + 1) for x, y in ranges]
a = b = 0
for r in ranges:
    for n in r:
        s = str(n)
        w = len(s) // 2
        if s == s[:w] * 2:
            a += n
            b += n
        else:
            for i in range(1, w + 1):
                if s == s[:i] * (len(s) // i):
                    b += n
                    break
print("answer_a:", a)
print("answer_b:", b)



"""
--- Day 14: One-Time Pad ---
https://adventofcode.com/2016/day/14
"""
import re
from _md5 import md5
from collections import defaultdict

from aocd import data


def search(data, stretch):
    template = data.encode() + b"%d"
    keys = []
    triples = defaultdict(list)
    pat3 = re.compile(r"(.)\1{2}")
    pat5 = re.compile(r"(.)\1{4}")
    i = 0
    stop = None
    while stop is None or i < stop:
        s = template % i
        for _ in range(stretch):
            s = md5(s).hexdigest().encode()
        s = s.decode()
        triple = pat3.search(s)
        if triple is not None:
            for quintuple in pat5.findall(s):
                keys.extend([x for x in triples[quintuple] if i - x <= 1000])
                triples[quintuple].clear()  # avoid to count same key twice
                if stop is None and len(keys) >= 64:
                    stop = i + 1001
            triples[triple.group()[0]].append(i)
        i += 1
    return sorted(keys)[63]


print("answer_a:", search(data, stretch=1))
print("answer_b:", search(data, stretch=2017))



"""
--- Day 11: Radioisotope Thermoelectric Generators ---
https://adventofcode.com/2016/day/11
"""
from itertools import combinations
from itertools import product

from aocd import data

from aoc_wim.search import AStar


def parsed(data):
    """
    state vector structure:
    (elevator, chip1, gen1, chip2, gen2 ... chipN, genN)
    """
    chips = {}
    generators = {}
    for line_no, line in enumerate(data.splitlines(), 1):
        words = line.split()
        for i, word in enumerate(words, -1):
            if word.startswith("generator"):
                generator = words[i]
                generators[generator] = line_no
            elif word.startswith("microchip"):
                chip, compatible = words[i].split("-")
                chips[chip] = line_no
    assert sorted(chips) == sorted(generators), "chip and generators mismatched"
    pairs = [(chips[k], generators[k]) for k in chips]
    state0 = sum(pairs, (1,))
    assert is_valid(state0), "initial state vector is invalid"
    return state0


def is_valid(state):
    chips = state[1::2]
    generators = state[2::2]
    for chip, generator in zip(chips, generators):
        if chip != generator:
            # if the chip and the matching generator are on different floors
            if chip in generators:
                # and there is some other generator on the same floor as the chip
                return False  # then this chip gets fried
    return True


class Q11AStar(AStar):
    def __init__(self, data, part="a"):
        state0 = parsed(data)
        if part == "b":
            state0 += (1,) * 4
        target = (4,) * len(state0)
        AStar.__init__(self, state0, target)

    def heuristic(self, state0, state1):
        return sum([abs(a - b) for (a, b) in zip(state0, state1)])

    cost = heuristic

    def adjacent(self, state):
        # indices of stuff on the same floor as the elevator
        idx = [i for i in range(1, len(state)) if state[0] == state[i]]
        # we can take 1 or 2 things with us when changing floors
        idx_choices = [*combinations(idx, 1), *combinations(idx, 2)]
        # we can go up or down a floor, unless on ground (1) or top (4)
        s0s = {1: (2,), 4: (3,)}.get(state[0], (state[0] - 1, state[0] + 1))
        for s0, idx in product(s0s, idx_choices):
            next_state = [s0] + list(state[1:])
            for i in idx:
                next_state[i] = s0
            pairs = zip(next_state[1::2], next_state[2::2])
            next_state = sum(sorted(pairs), (s0,))
            if is_valid(next_state):
                yield next_state


a = Q11AStar(data, part="a")
a.run()
print("answer_a:", a.path_length)

b = Q11AStar(data, part="b")
b.run()
print("answer_b:", b.path_length)



"""
--- Day 21: Fractal Art ---
https://adventofcode.com/2017/day/21
"""
import numpy as np
from aocd import data
from aocd import extra


def s2a(s):
    return np.array([[{".": 0, "#": 1}[c] for c in row] for row in s.split("/")])


def a2s(a):
    return "/".join("".join(".#"[n] for n in row) for row in a)


def parsed(data):
    rules = {}
    for line in data.splitlines():
        a, b = line.split(" => ")
        a, b = s2a(a), s2a(b)
        for i in range(4):
            rules[a2s(a)] = rules[a2s(np.flipud(a))] = rules[a2s(np.fliplr(a))] = b
            a = np.rot90(a)
    return rules


def evolve(data, iterations):
    rules = parsed(data)
    grid = s2a(".#./..#/###")
    for i in range(iterations):
        n = len(grid)
        s = 3 if n % 2 else 2
        n //= s
        rows = []
        for row in range(n):
            r = []
            for col in range(n):
                c = grid[row * s : (row + 1) * s, col * s : (col + 1) * s]
                r.append(rules[a2s(c)])
            r = np.hstack(r)
            rows.append(r)
        grid = np.vstack(rows)
    return grid.sum()


n = extra.get("iterations", 5)
print("answer_a:", evolve(data, n))
if not extra:
    print("answer_b:", evolve(data, 18))



"""
--- Day 10: Pipe Maze ---
https://adventofcode.com/2023/day/10
"""
import networkx as nx
from aocd import data

from aoc_wim.zgrid import ZGrid

data = data.translate(str.maketrans("-|F7LJ", "─│╭╮╰╯"))
grid = ZGrid(data)
s = grid.z("S")
graph = nx.Graph()
graph.add_nodes_from(grid)
horizontals = """\
── ─╯ ─╮ ╰─ ╰╯ ╰╮ ╭─ ╭╯ ╭╮
"""
verticals = """\
│ │ │ ╮ ╮ ╮ ╭ ╭ ╭
│ ╯ ╰ │ ╯ ╰ │ ╯ ╰
"""
verticals = [a + b for a, b in zip(*verticals.splitlines())][::2]

for z, v in grid.scan():
    vr = grid.get(z + 1, ".")
    vd = grid.get(z + 1j, ".")
    if v + vr in horizontals:
        graph.add_edge(z, z + 1)
    if v + vd in verticals:
        graph.add_edge(z, z + 1j)

dz_L = zip((-1j, 1, 1j, -1), "│╭╮ ─╮╯ │╰╯ ─╭╰".split())
s_join = {dz: grid.get(s + dz, ".") in L for dz, L in dz_L}
for dz, join in s_join.items():
    if join:
        graph.add_edge(s, s + dz)
match *s_join.values(),:
    case 0, 0, 1, 1: grid[s] = "╮"
    case 0, 1, 0, 1: grid[s] = "─"
    case 0, 1, 1, 0: grid[s] = "╭"
    case 1, 0, 0, 1: grid[s] = "╯"
    case 1, 0, 1, 0: grid[s] = "│"
    case 1, 1, 0, 0: grid[s] = "╰"

[loop] = [c for c in nx.connected_components(graph) if s in c]
print("answer_a:", len(loop) // 2)

interior = []
inside = False
for z, v in grid.scan():
    if z in loop and v in "│╭╮":
        inside = not inside
    if z not in loop and inside:
        interior.append(z)

print("answer_b:", len(interior))

# from termcolor import colored
# o = {z: colored(grid[z], "red") for z in loop}
# o[s] = colored("S", "blue")
# o |= {z: colored(grid[z], "green") for z in interior}
# grid.draw(overlay=o, hide_ax=True)



import logging
import string
from collections import ChainMap
from collections import deque

import networkx as nx
import numpy as np


log = logging.getLogger(__name__)


def manhattan_distance(z1, z0=0):
    dz = z1 - z0
    return abs(int(dz.real)) + abs(int(dz.imag))


def manhattan_ball(z0=0, r=1, full=False):
    """circle/ball centered on z0 with radius r (using manhattan distance)"""
    result = []
    if full or r == 0:
        result.append(0)
        rs = range(1, r + 1)
    else:
        rs = [r]
    for r in rs:
        for x in range(r):
            result.append(complex(x, x - r))
        for y in range(r):
            result.append(complex(r - y, y))
        for x in range(r):
            result.append(complex(-x, r - x))
        for y in range(r):
            result.append(complex(y - r, -y))
    result = [z + z0 for z in result]
    return result


def hexagonal_distance(z1, z0=0):
    dz = z1 - z0
    return int(max(abs(dz.real), abs(dz.imag), abs(dz.real - dz.imag)))


class ZDict(dict):

    def __init__(self, func):
        self.func = func

    def __missing__(self, z):
        if not isinstance(z, (int, complex, float)):
            log.error("ZDict does not support key %r", z)
            raise NotImplementedError
        self[z] = self.func(z)
        return self[z]


class ZGrid:

    _dzs = [-1j, 1, 1j, -1]
    dzs = ChainMap(
        dict(zip(_dzs, _dzs)),
        dict(zip("^>v<", _dzs)),
        dict(zip("up right down left".split(), _dzs)),
        dict(zip("up right down left".upper().split(), _dzs)),
        dict(zip("URDL", _dzs)),
        dict(zip("NESW", _dzs)),
        dict(zip("forward backward".split(), _dzs[1::2])),
    )
    U = N = up = north = -1j
    R = E = right = east = 1
    D = S = down = south = 1j
    L = W = left = west = -1
    UR = NE = N + E
    UL = NW = N + W
    DL = SW = S + W
    DR = SE = S + E

    turn_right = turnR = 1j
    turn_left = turnL = -1j
    turn_around = -1

    def __init__(self, initial_data=None, transform=None):
        if isinstance(transform, dict):
            t = transform.copy()
            def transform(k):
                return t.get(k, k)
        self.d = d = {}
        if initial_data is not None:
            if isinstance(initial_data, ZGrid):
                self.d = initial_data.d.copy()
            elif isinstance(initial_data, str):
                for row, line in enumerate(initial_data.splitlines()):
                    for col, char in enumerate(line):
                        if transform is not None:
                            char = transform(char)
                        if char is not None:
                            d[col + row*1j] = char
            elif callable(initial_data):
                self.d = ZDict(func=initial_data)
            elif isinstance(initial_data, dict):
                k, _ = _, initial_data[k] = initial_data.popitem()
                if isinstance(k, tuple) and len(k) == 2:
                    k0, k1 = k
                    if isinstance(k0, int) and isinstance(k1, int):
                        initial_data = {k[1] + 1j*k[0]: v for k, v in initial_data.items()}
                self.d = initial_data
            elif isinstance(initial_data, np.ndarray) and initial_data.ndim == 2:
                for row, r in enumerate(initial_data):
                    for col, char in enumerate(r):
                        if transform is not None:
                            char = transform(char)
                        d[col + row*1j] = char
            else:
                raise NotImplementedError(f"{type(initial_data)=}")

    @classmethod
    def fromempty(cls, width, height, glyph="."):
        return cls({z: glyph for z in zrange(0, complex(width, height))})

    def offset(self, dz):
        self.d = {(z + dz): v for z, v in self.items()}

    def __setitem__(self, key, value):
        self.d[key] = value

    def __getitem__(self, key):
        return self.d[key]

    def __delitem__(self, key):
        del self.d[key]

    def __contains__(self, item):
        return item in self.d

    def __iter__(self):
        return iter(self.d)

    def scan(self):
        for k in sorted(self.keys(), key=lambda v: (v.imag, v.real)):
            yield k, self.d[k]

    def __len__(self):
        return len(self.d)

    def items(self):
        return self.d.items()

    def keys(self):
        return self.d.keys()

    def values(self):
        return self.d.values()

    def count(self, val):
        return sum([1 for v in self.values() if v == val])

    def count_near(self, z0, val, *, n=4, default=None):
        vals = [self.get(z, default) for z in self.near(z0, n=n)]
        result = vals.count(val)
        return result

    def get(self, k, default=None):
        return self.d.get(k, default)

    def z(self, val, first=True):
        zs = []
        for k, v in self.items():
            if v == val:
                if first:
                    return k
                zs.append(k)
        return zs

    def remove(self, val):
        keep = {z: v for z, v in self.items() if v != val}
        n_removed = len(self) - len(keep)
        self.d.clear()
        self.d.update(keep)
        return n_removed

    zball = staticmethod(manhattan_ball)

    @staticmethod
    def near(z, n=4):
        if n == 4:
            return [
                        z-1j,
                z-1,             z+1,
                        z+1j,
            ]
        if n == 5:
            return [
                        z-1j,
                z-1,    z,       z+1,
                        z+1j,
            ]
        elif n == 6:  # hexgrid w/ skewed coordinate system
            return [
                z-1-1j, z-1j,
                z-1,             z+1,
                        z+1j, z+1+1j,
            ]
        elif n == 7:
            return [
                z-1-1j, z-1j,
                z-1,    z,       z+1,
                        z+1j, z+1+1j,
            ]
        elif n == 8:
            return [
                z-1-1j, z-1j, z+1-1j,
                z-1,             z+1,
                z-1+1j, z+1j, z+1+1j,
            ]
        elif n == 9:
            return [
                z-1-1j, z-1j, z+1-1j,
                z-1,    z,       z+1,
                z-1+1j, z+1j, z+1+1j,
            ]

    def draw(self, overlay=None, window=None, clear=False, pretty=False, transform=None, title="", flip=None, empty_glyph=None, hide_ax=False):
        if window is None:
            d = self.d
        else:
            if isinstance(window, complex):
                window = zrange(window + 1 + 1j)
            d = {z: self[z] for z in window}
        if overlay is not None:
            d = {**self.d, **overlay}
        dump_grid(d, clear=clear, pretty=pretty, transform=transform, title=title, flip=flip, empty_glyph=empty_glyph, hide_ax=hide_ax)

    # TODO:
    #  hexgrid compass overlay ✶
    #  odd-r / even-r etc modes?
    #  zoom / rotate / reflect
    #  make sure axes line up with labels
    #  better x-axis values
    def draw_hex(self, clear=False, glyph=2, labels=False, orientation="V", title=""):
        cell = HexCell(hex_glyph_gen(glyph, orientation=orientation))
        plane = Plane()
        label = ""
        for z, val in self.items():
            row, col = transform(z, orientation)
            if labels:
                label = f"{col},{row}"
            draw_cell(plane, cell, row, col, val, label=label)
        plane.draw(clear=clear, xscale=cell.dx, yscale=cell.dy, cellwidth=cell.w, title=title)
        return plane

    def translate(self, table):
        for z in self.d:
            if self.d[z] in table:
                self.d[z] = table[self.d[z]]

    def __array__(self, dtype=None, copy=True):
        """makes np.array(zgrid) work"""
        if not copy:
            raise NotImplementedError
        zs = np.array(list(self.d))
        xs = zs.real.astype(int)
        ys = zs.imag.astype(int)
        vs = np.array(list(self.d.values()))
        w = np.ptp(xs) + 1
        h = np.ptp(ys) + 1
        fill = "." if vs.dtype == "U1" else 0
        full = np.full((h, w), fill_value=fill, dtype=vs.dtype)
        full[ys - ys.min(), xs - xs.min()] = vs
        if dtype is not None:
            full = full.astype(dtype)
        return full

    def graph(self, on=".", extra=()):
        """connected components"""
        node_glyphs = {on}.union(extra)
        g = nx.Graph()
        g.extra = {}
        for pos, glyph in self.d.items():
            if glyph in node_glyphs:
                g.add_node(pos)
                if glyph != on:
                    g.extra[glyph] = pos
                right = pos + 1
                down = pos + 1j
                if self.d.get(right) in node_glyphs:
                    g.add_edge(pos, right)
                if self.d.get(down) in node_glyphs:
                    g.add_edge(pos, down)
        return g

    def path(self, z, z0=0):
        g = self.graph()
        return nx.shortest_path(g, z0, z)

    def path_length(self, z, z0=0):
        g = self.graph()
        return nx.shortest_path_length(g, z0, z)

    def draw_path(self, z, z0=0, glyph="x", clear=False, pretty=True):
        path = self.path(z, z0)
        overlay = {}.fromkeys(path, glyph)
        overlay[path[0]] = "O"
        overlay[path[-1]] = "T"
        self.draw(overlay=overlay, clear=clear, pretty=pretty)

    def bfs(self, target=None, z0=0, max_depth=None, on="."):
        """returns a dict of connected nodes vs depth up to max_depth"""
        g0 = self[z0]
        if g0 != on:
            log.error("Expected initial glyph %r, got %r", on, g0)
            raise NotImplementedError
        seen = {}
        queue = deque([(z0, 0)])
        while queue:
            z0, depth = queue.popleft()
            if max_depth is not None and depth > max_depth:
                return seen
            if z0 not in seen:
                seen[z0] = depth
                if target is not None and z0 == target:
                    return seen
                for z in self.near(z0):
                    if z in seen:
                        continue
                    try:
                        g = self[z]
                    except KeyError:
                        continue
                    if g == on:
                        queue.append((z, depth + 1))
        return seen

    @property
    def top_left(self):
        return min(self.d, key=lambda z: (z.imag, z.real))

    @property
    def left_bottom(self):
        return min(self.d, key=lambda z: (z.real, -z.imag))

    @property
    def bottom_right(self):
        return max(self.d, key=lambda z: (z.imag, z.real))

    @property
    def right_top(self):
        return max(self.d, key=lambda z: (z.real, -z.imag))

    @property
    def width(self):
        if not self.d:
            return 0
        return int(self.right_top.real - self.left_bottom.real) + 1

    @property
    def height(self):
        if not self.d:
            return 0
        return int(self.bottom_right.imag - self.top_left.imag) + 1

    @property
    def area(self):
        return self.height * self.width

    def aabb(self, pad=0):
        """axis aligned bounding box"""
        x0 = min(z.real for z in self)
        x1 = max(z.real for z in self)
        y0 = min(z.imag for z in self)
        y1 = max(z.imag for z in self)
        top_left = complex(x0, y0)
        bottom_right = complex(x1, y1)
        return zrange(top_left - (1 + 1j)*pad, bottom_right + (1 + 1j)*(pad + 1))

    def update(self, other):
        if isinstance(other, ZGrid):
            other = other.d
        self.d.update(other)


def dump_grid(g, clear=False, pretty=True, transform=None, title="", flip=None, empty_glyph=None, hide_ax=False):
    if transform is None:
        transform = {
            "#": "⬛",
            ".": "  ",
            "O": "🤖",
            "T": "🥇",
            "x": "👣",
            ">": "➡️ ",
            "<": "⬅️ ",
            "^": "⬆️ ",
            "v": "⬇️ ",
            "@": "@️ ",
            0: "  ",
            1: "⬛",
        }
    if pretty:
        transform.update({x: x + " " for x in string.ascii_letters if x not in transform})
        if empty_glyph is None:
            empty_glyph = "  "
    if empty_glyph is None:
        empty_glyph = "."
    print()
    xs = [int(z.real) for z in g]
    ys = [int(z.imag) for z in g]
    cols = range(min(xs), max(xs) + 1)
    rows = range(min(ys), max(ys) + 1)
    W = len(cols)
    if pretty:
        W *= 2
    if clear:
        print("\033c")
    if pretty:
        print(" "*5 + "┌" + title.center(W, "─") + "┐")
    elif title:
        print(" "*5 + title.center(W))
    if "x" in (flip or ""):
        cols = cols[::-1]
    if "y" in (flip or ""):
        rows = rows[::-1]
    for row in rows:
        if not hide_ax:
            print(f"{row:>5d}", end="")
        line = []
        if pretty:
            line.append("│")
        else:
            line.append(" ")
        for col in cols:
            glyph = g.get(col + row * 1j, empty_glyph)
            if pretty:
                glyph = transform.get(glyph, glyph)
            line.append(str(glyph))
        if pretty:
            line.append("│")
        print("".join(line))
    if pretty:
        print(" "*5 + "└" + "─"*W + "┘")
    if not hide_ax:
        footer_left = f"{cols[0]}".ljust(W)
        footer_center = f"{cols[len(cols)//2]}".center(W+1)
        footer_right = f"{cols[-1]}".rjust(W)
        zf = zip(footer_left, footer_center, footer_right)
        footer = [next((x for x in iter([l, c, r]) if x != " "), " ") for (l, c, r) in zf]
        footer = "".join(footer)
        print(" " * 6 + footer)
        print()


def array2txt(a):
    a = a.astype(str)
    lines = ["".join(row) for row in a]
    txt = "\n".join(lines)
    return txt


def zrange(*args):
    # a 2D rectangle of complex points
    if len(args) == 1:
        start = 0
        (stop,) = args
        step = 1 + 1j
    elif len(args) == 2:
        start, stop = args
        step = 1 + 1j
    elif len(args) == 3:
        start, stop, step = args
    else:
        raise TypeError(f"zrange expected 1-3 arguments, got {len(args)}")
    xs = range(int(start.real), int(stop.real), int(step.real))
    ys = range(int(start.imag), int(stop.imag), int(step.imag))
    return [complex(x, y) for y in ys for x in xs]


# axes unit vectors for hexagonal grid with skewed coordinate system - flat-topped
hexV = dict(zip("n ne nw se sw s".split(), ZGrid().near(0, n=6)))

# hexgrid - pointy-topped
hexH = dict(zip("w nw sw ne se e".split(), ZGrid().near(0, n=6)))


def hex_glyph_gen(n, fill=".", orientation="V"):
    if orientation not in ["V", "H"]:
        # V is hexgrid with a North-South axis, i.e. flat-topped hexagons
        # H is hexgrid with an East-West axis, i.e. pointy-topped hexagons
        raise ValueError("Orientation must be 'V' or 'H'")
    if n == 0:
        return {"V": "⬣", "H": "⬢"}[orientation]
    if orientation == "H":
        # TODO: ascii-art for pointy-topped hexes
        #  _.-''-._
        # |        |
        # |        |
        #  '-.,,.-'
        #
        #      _.-''-._
        #  _.-'        '-._
        # |                |
        # |                |
        # |                |
        # |                |
        #  '-._        _.-'
        #      '-.,,.-'
        #
        #  .-''-.
        # /      \
        #|        |
        # \      /
        #  `-..-'
        #
        #  _.-´`-._
        # |        |     warning: ´ is non-ascii
        # |        |     ⹁ too
        #  '-.⹁,.-'

        # '' - -~~.., , __

        # _,.-'"^`

        # __,, ..--"" ^ ^ ^

        raise NotImplementedError
    first = " "*n + "__"*n
    last = " "*(n - 1) + "\\" + "__"*n + "/"
    lines = [first, last]
    for i in range(n):
        center = fill*2*(n + i)
        left = " "*(n - 1 - i)
        top = left + "/" + center + "\\"
        bottom = left + "\\" + center + "/"
        mid = len(lines) // 2
        lines[mid:mid] = [top, bottom]
    del lines[-2]
    return "\n".join(lines)


class HexCell:

    def __init__(self, glyph, dy=None, dx=None, fill="."):
        if isinstance(glyph, int):
            glyph = hex_glyph_gen(glyph, fill=fill)
        self.glyph = glyph.strip("\n")
        lines = self.glyph.splitlines()
        self.h = len(lines)
        self.w = max([len(x) for x in lines])
        if dy is None:
            for dy, line in enumerate(lines):
                if line.strip().startswith("\\"):
                    dy -= 1
                    break
        self.dy = dy or 1
        if dx is None and len(lines) > 1:
            for dx, char in enumerate(lines[1]):
                if char == "\\":
                    break
        self.dx = dx or 1
        self.blanked = self.glyph.replace(fill, " ")
        if self.glyph == "⬣":
            self.blanked = "⎔"
            self.dx = 2
        if self.glyph == "⬢":
            self.blanked = "⬡"
            self.dx = 1


class Plane:
    """Unbounded 2D rectangular space of values that automatically grows as needed"""
    def __init__(self, left=0, right=0, top=0, bottom=0, fill=" "):
        if left > right or top > bottom:
            raise ValueError(f"invalid initial: {left},{right},{top},{bottom}")
        self.left = left
        self.right = right
        self.top = top
        self.bottom = bottom
        self.lines = {}
        self.fill = fill
        w = self.width
        self.lines = {y: deque([fill]*w) for y in range(self.top, self.bottom + 1)}

    def __getitem__(self, item):
        row, col = item
        col -= self.left
        return self.lines[row][col]

    def __setitem__(self, item, val):
        row, col = item
        while row < self.top:
            self.top -= 1
            self.lines[self.top] = deque([self.fill]*self.width)
        while row > self.bottom:
            self.bottom += 1
            self.lines[self.bottom] = deque([self.fill]*self.width)
        if col < self.left:
            grow = [self.fill] * (self.left - col)
            for line in self.lines.values():
                line.extendleft(grow)
            self.left = col
        if col > self.right:
            grow = [self.fill] * (col - self.right)
            for line in self.lines.values():
                line.extend(grow)
            self.right = col
        col -= self.left
        self.lines[row][col] = val

    @property
    def width(self):
        return self.right - self.left + 1

    @property
    def height(self):
        return self.bottom - self.top + 1

    def draw(self, clear=True, xscale=1, yscale=1, cellwidth=4, title=""):
        if clear:
            print("\033c")
        y_pad = max([len(str(r//yscale)) for r in self.lines])
        print(" "*y_pad + "┌" + title.center(self.width, "─") + "┐")
        prev_tick = None
        for row in range(self.top, self.bottom + 1):
            line = self.lines[row]
            tick = str(row//yscale)
            if tick == prev_tick:
                # don't write it again
                tick = ""
            else:
                prev_tick = tick
            print(tick.rjust(y_pad) + "│" + "".join(line) + "│")
        print(" "*y_pad + "└" + "─"*self.width + "┘")
        footer = [" "] * self.width
        for i, char in enumerate(str((self.left + cellwidth//2)//xscale)):
            footer[i] = char
        for i, char in enumerate(reversed(str(self.right//xscale))):
            footer[~i] = char
        if self.left <= 0 <= self.right:
            i0 = -self.left - 1
            footer[i0] = "0"
        else:
            mid = self.left + self.width//2
            imid = sorted(self.lines).index(mid)
            mid_label = str(mid)
            imidl = imid - 1
            imidr = imid + len(mid_label) + 1
            if imidl > 0 and footer[imidl] == " ":
                if imidr < len(footer) and footer[imidr] == " ":
                    for char in mid_label:
                        footer[imid] = char
                        imid += 1
        print(" "*(y_pad + 1) + "".join(footer))
        print()


def draw_cell(plane, cell, r, c, val=False, label=""):
    glyph = cell.glyph if val else cell.blanked
    r *= cell.dy
    c *= cell.dx
    for i, row in enumerate(glyph.splitlines(), start=-cell.dy):
        for j, char in enumerate(row, start=(-cell.w//2)):
            if char == " ":
                continue
            plane[r + i, c + j] = char
    if label and cell.w > 4:
        c = c - cell.w//2 + 1
        for char in label:
            plane[r, c] = char
            c += 1


def transform(z, orientation="V"):
    # change of coordinate system: axial to offset
    real, imag = int(z.real), int(z.imag)
    if orientation == "V":
        # 1  --> (1, 1)
        # 1j --> (1, -1)
        row = real + imag
        col = real - imag
    elif orientation == "H":
        # 1  --> (-1, 1)
        # 1j --> (1, 1)
        row = imag - real
        col = imag + real
    else:
        raise ValueError("Orientation must be 'V' or 'H'")
    return row, col


def zline(p1, p2):
    # points between p1 and p2 inclusive of both endpoints
    match p1, p2:
        case int(), int() if p1 <= p2:
            return range(p1, p2 + 1)
        case int(), int() if p1 > p2:
            return range(p1, p2 - 1, -1)
        case (int() | float() | complex()), (int() | float() | complex()):
            for n in p1, p2:
                if isinstance(n, (float, complex)):
                    if not n.imag.is_integer() or not n.real.is_integer():
                        raise NotImplementedError
            d = p2 - p1
            if d == 0:
                return [p1]
            if d.real == 0:
                dz = -1j if d.imag < 0 else 1j
            elif d.imag == 0:
                dz = -1 if d.real < 0 else 1
            elif abs(d.real) == abs(d.imag):
                dz = d / abs(d.real)
            else:
                raise NotImplementedError
            result = [p1]
            while p1 != p2:
                p1 += dz
                result.append(p1)
            return result
        case _:
            raise NotImplementedError



"""
--- Day 17: Chronospatial Computer ---
https://adventofcode.com/2024/day/17
"""
import re

from aocd import data


class Comp:
    combo = {4: "A", 5: "B", 6: "C"}

    def __init__(self, A, B, C, prog=()):
        self.output = []
        self.reset(A, B, C, prog)

    def __getitem__(self, k):
        k = self.combo.get(k, k)
        return self.data.get(k, k)

    def __setitem__(self, k, v):
        k = self.combo.get(k, k)
        self.data[k] = v

    def reset(self, A=0, B=0, C=0, prog=()):
        self.data = {"A": A, "B": B, "C": C}
        self.prog = prog
        self.ip = 0
        self.output.clear()

    def adv(self, x):
        self["A"] >>= self[x]

    def bxl(self, x):
        self["B"] ^= x

    def bst(self, x):
        self["B"] = self[x] & 0o7

    def jnz(self, x):
        if self["A"]:
            self.ip = x - 2

    def bxc(self, x):
        self["B"] ^= self["C"]

    def out(self, x):
        self.output.append(self[x] & 0o7)

    def bdv(self, x):
        self["B"] = self["A"] >> self[x]

    def cdv(self, x):
        self["C"] = self["A"] >> self[x]

    ops = [adv, bxl, bst, jnz, bxc, out, bdv, cdv]

    def run(self):
        while self.ip < len(self.prog):
            opcode, operand = self.prog[self.ip:self.ip+2]
            func = self.ops[opcode]
            func(self, operand)
            self.ip += 2
        return ",".join(map(str, self.output))


A, B, C, *prog = map(int, re.findall(r"\d+", data))
comp = Comp(A, B, C, prog=prog)
print("answer_a:", comp.run())

quines = []
stack = [0]
while stack:
    A0 = stack.pop()
    for n in range(0o10):
        A = (A0 << 3) + n
        comp.reset(A=A, prog=prog)
        comp.run()
        if comp.output == prog[-len(comp.output):] and A:
            stack.append(A)
        if comp.output == prog:
            quines.append(A)

print("answer b:", min(quines, default=None))



"""
--- Day 19: Aplenty ---
https://adventofcode.com/2023/day/19
"""
from dataclasses import dataclass
from dataclasses import replace
from math import prod

from aocd import data


workflows_raw, parts_raw = data.split("\n\n")

workflows = {}
for line in workflows_raw.splitlines():
    pk, w = line.rstrip("}").split("{")
    wf = []
    *rules, else_dst = w.split(",")
    for rule in rules:
        rule, dst = rule.split(":")
        op = "<" if "<" in rule else ">"
        ax, val = rule.split(op)
        wf.append((ax, op, int(val), dst))
    wf.append(else_dst)
    workflows[pk] = wf

parts = []
for r in parts_raw.splitlines():
    # just can't bring myself to use eval/exec
    part = {}
    for x_n in r.strip("{}").split(","):
        x, n = x_n.split("=")
        part[x] = int(n)
    parts.append(part)

a = 0
for part in parts:
    node = "in"
    while node not in "AR":
        *rules, else_dst = workflows[node]
        for rule in rules:
            ax, op, val, dst = rule
            result = part[ax] < val if op == "<" else part[ax] > val
            if result:
                node = dst
                break
        else:
            node = else_dst
    if node == "A":
        a += sum(part.values())
print("answer_a:", a)


@dataclass
class Block:
    x0: int = 1
    m0: int = 1
    a0: int = 1
    s0: int = 1
    x1: int = 4001
    m1: int = 4001
    a1: int = 4001
    s1: int = 4001

    @property
    def volume(self):
        return prod([
            self.x1 - self.x0,
            self.m1 - self.m0,
            self.a1 - self.a0,
            self.s1 - self.s0,
        ])

    def split(self, ax, op, val):
        if op == ">":
            return self.split(ax, "<", val + 1)[::-1]
        block_true = replace(self, **{f"{ax}1": val})
        block_false = replace(self, **{f"{ax}0": val})
        return block_true, block_false


b = 0
stack = [(Block(), "in")]
while stack:
    x, node = stack.pop()
    if node == "A":
        b += x.volume
        continue
    if node == "R":
        continue
    *rules, else_dst = workflows[node]
    for ax, op, val, dst in rules:
        xt, x = x.split(ax, op, val)
        stack.append((xt, dst))
    stack.append((x, else_dst))
print("answer_b:", b)



"""
--- Day 5: If You Give A Seed A Fertilizer ---
https://adventofcode.com/2023/day/5
"""
from aocd import data


chunks = data.split("\n\n")
seeds, *chunks = chunks
seeds = [int(x) for x in seeds.removeprefix("seeds:").split()]
mm = []
for chunk in chunks:
    _, *ranges = chunk.splitlines()
    m = []
    for r in ranges:
        dest, source, length = map(int, r.split())
        delta = dest - source
        m.append([source, source + length, delta])
    m.sort()
    m_new = []
    # fill any gaps
    for m1, m2 in zip(m, m[1:]):
        m_new.append(m1)
        if m1[1] < m2[0]:
            m_new.append([m1[1], m2[0], 0])
    m_new.extend(m[-1:])
    m[:] = m_new
    mm.append(m)

s_prev = seeds[:]
for m in mm:
    s_new = []
    for s in s_prev:
        for left, right, delta in m:
            if left <= s < right:
                s_new.append(s + delta)
                break
        else:
            s_new.append(s)
    s_prev[:] = s_new

a = min(s_prev)
print("answer_a:", a)

r_prev = [(s0, s0 + s1) for s0, s1 in zip(seeds[0::2], seeds[1::2])]
for m in mm:
    cuts = [left for left, right, delta in m]
    cuts.append(m[-1][1])
    r_new = []
    r_cut = []
    for L, R in r_prev:
        inside = [L] + [c for c in cuts if L < c < R] + [R]
        r_cut.extend(zip(inside, inside[1:]))
    for left0, right0 in r_cut:
        for left, right, d in m:
            if left <= left0 < right0 <= right:
                break
        else:
            d = 0
        r_new.append((left0 + d, right0 + d))
    r_prev[:] = r_new

b = min(left for left, right in r_prev)
print("answer_b:", b)



"""
--- Day 9: Disk Fragmenter ---
https://adventofcode.com/2024/day/9
"""
from collections import defaultdict
from collections import deque
from dataclasses import dataclass
from heapq import heappop
from heapq import heappush

from aocd import data


@dataclass
class Mem:
    id: int
    pos: int
    size: int

    @property
    def checksum(self):
        return self.id * (2 * self.pos + self.size - 1) * self.size // 2

    def __lt__(self, other):
        if not isinstance(other, Mem):
            return NotImplemented
        return self.pos < other.pos


def parsed(data):
    disk = deque(), deque()
    pos = 0
    for i, size in enumerate(map(int, data)):
        fid, i = divmod(i, 2)
        if size:
            disk[i].append(Mem(fid, pos, size))
        pos += size
    return disk


def defrag_a(data):
    files, free = parsed(data)
    while free[0].pos <= files[-1].pos:
        s = free.popleft()
        f = files.pop()
        if f.size <= s.size:
            f.pos = s.pos
            files.appendleft(f)
            s.size -= f.size
            s.pos += f.size
            if s.size:
                free.appendleft(s)
        else:
            files.append(Mem(f.id, f.pos, f.size - s.size))
            f.pos = s.pos
            f.size = s.size
            files.appendleft(f)
    return sum(f.checksum for f in files)


def defrag_b(data):
    files, free = parsed(data)
    heaps = defaultdict(list)
    for f in free:
        heappush(heaps[f.size], Mem(0, f.pos, f.size))

    for f in reversed(files):
        slots = [h for s, h in heaps.items() if s >= f.size and h and h[0].pos < f.pos]
        if not slots:
            continue
        heap = min(slots, key=lambda h: h[0].pos)
        s = heappop(heap)
        f.pos = s.pos
        s.size -= f.size
        s.pos += f.size
        heappush(heaps[s.size], s)
    return sum(f.checksum for f in files)


print("answer_a:", defrag_a(data))
print("answer_b:", defrag_b(data))



"""
--- Day 3: Lobby ---
https://adventofcode.com/2025/day/3
"""

from aocd import data


def max_jolts(line, d=2):
    s = []
    i = 0
    while d:
        m = max(line[i : 1 - d or None])
        s.append(m)
        i = line.index(m, i) + 1
        d -= 1
    return int("".join(s))


lines = data.split()
print("answer_a:", sum(max_jolts(x, d=2) for x in lines))
print("answer_b:", sum(max_jolts(x, d=12) for x in lines))



"""
--- Day 2: Bathroom Security ---
https://adventofcode.com/2016/day/2
"""
from aocd import data

from aoc_wim.zgrid import ZGrid


keypad_a = """\
123
456
789
"""


keypad_b = """\
  1
 234
56789
 ABC
  D
"""


def decode(keypad):
    keypad = ZGrid(keypad)
    z = keypad.z("5")
    code = ""
    for line in data.splitlines():
        for direction in line:
            dz = getattr(ZGrid, direction)
            if keypad.get(z + dz, " ").strip():
                z += dz
        code += keypad[z]
    return code


print("answer_a:", decode(keypad_a))
print("answer_b:", decode(keypad_b))



"""
--- Day 16: Aunt Sue ---
https://adventofcode.com/2015/day/16
"""
from aocd import data


message = """\
children: 3
cats: 7
samoyeds: 2
pomeranians: 3
akitas: 0
vizslas: 0
goldfish: 5
trees: 3
cars: 2
perfumes: 1
"""


d_message = {k: int(v) for k, v in (pair.split(": ") for pair in message.splitlines())}

sues = {}
for line in data.splitlines():
    sue, stats = line.split(": ", 1)
    stats = {k: int(v) for k, v in (pair.split(": ") for pair in stats.split(", "))}
    sues[sue] = stats


def find_sue(**kwargs):
    for sue, stats in sues.items():
        common_keys = d_message.keys() & stats.keys()
        if all(kwargs.get(k, int.__eq__)(stats[k], d_message[k]) for k in common_keys):
            return int(sue.split()[1])


a = find_sue()
b = find_sue(
    cats=int.__gt__,
    trees=int.__gt__,
    pomeranians=int.__lt__,
    goldfish=int.__lt__,
)

print("answer_a:", a)
print("answer_b:", b)
